from math import pi, sqrt
import numpy as np
from xespresso import Espresso
from ase.geometry import get_layers
import copy
import matplotlib.pyplot as plt
from ase.data.colors import jmol_colors
from ase.data import chemical_symbols

lspins = ['up', 'dw']
orbitals = ['s', 'p', 'd', 'f']


class DOS:
    def __init__(self, calc=None, label=None, prefix=None,
                 debug=False, colors=None, spin_orbit_coupling=False):
        """Electronic Density Of States object.

        calc: quantum espresso calculator object
        label: the directory and prefix

        dos_energies
        print(pdos_info) to see the detail
        pdos_kinds[species][channel][ncomponents]

        """
        if not calc and not label:
            raise ValueError('Please give one of them: calc or label')
        if label:
            calc = Espresso(label=label, prefix=prefix, debug=True)
        calc.read_results()
        self.debug = debug
        self.spin_orbit_coupling = spin_orbit_coupling
        if self.debug:
            print(calc.results)
        self.directory = calc.directory
        self.label = calc.label
        self.prefix = calc.prefix
        self.efermi = calc.get_fermi_level()
        self.atoms = calc.results['atoms']
        self.nspins = calc.get_number_of_spins()
        if self.nspins == 0:
            self.nspins = 1

    def read_dos(self):
        dos = np.loadtxt(self.directory + '/dos/%s.dos' % self.prefix)
        if self.nspins == 2:
            self.dos = [dos[:, 1], dos[:, 2]]
        else:
            self.dos = [dos[:, 1]]
        self.dos_energies = dos[:, 0] - self.efermi
        # self.dos_energies = self.dos_energies.reshape((self.dos_energies.shape[0], 1))

    def read_pdos_info(self, ):
        '''
        read pdos information from output of projwfc
        state #   1: atom   1 (Al ), wfc  1 (l=0 m= 1)
        state #   2: atom   1 (Al ), wfc  2 (l=1 m= 1)
        state #   3: atom   1 (Al ), wfc  2 (l=1 m= 2)
        state #   4: atom   1 (Al ), wfc  2 (l=1 m= 3)
        state #   5: atom   2 (Al ), wfc  1 (l=0 m= 1)
        state #   6: atom   2 (Al ), wfc  2 (l=1 m= 1)
        state #   7: atom   2 (Al ), wfc  2 (l=1 m= 2)
        ...

        reg
        ['state', '1', 'atom', '1', 'Al', 'wfc', '1', 'l', '0', 'm', '1']
        Output: 
        post_info: dict,
            e.g. {'Al': {'iatom': [1, 2, 3, 4], 'istate': {1: 0, 2: 1}}}
        '''
        # import re
        # reg = re.compile('\w+')
        #
        kinds = []
        pdos_info = {}
        with open(self.directory + '/projwfc/%s.projwfco' % self.prefix, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'state #' in line:
                    for s in [',', '(', ')', '=', '#']:
                        line = line.replace(s, ' ')
                    data = line.split()
                    iatom = int(data[3])
                    kind = data[4]
                    istate = int(data[6])
                    if self.spin_orbit_coupling:
                        l = int(data[10])
                        j = float(data[8])
                    else:
                        l = int(data[8])
                        j = None
                    if kind not in kinds:
                        kinds.append(kind)
                        pdos_info[kind] = {}
                        pdos_info[kind]['iatom'] = []
                        pdos_info[kind]['istate'] = {}
                    if iatom not in pdos_info[kind]['iatom']:
                        pdos_info[kind]['iatom'].append(iatom)
                    if istate not in pdos_info[kind]['istate']:
                        if j:
                            pdos_info[kind]['istate'][istate] = [l, '_j%s' % j]
                        else:
                            pdos_info[kind]['istate'][istate] = [l]
        self.kinds = kinds
        self.pdos_info = pdos_info
        if self.debug:
            print(self.pdos_info)

    def read_pdos(self):
        '''
        projwfc
        '''
        self.read_pdos_info()
        # read in total density of states
        dos = np.loadtxt(self.directory + '/projwfc/%s.pdos_tot' % self.prefix)
        if self.nspins == 2:
            self.pdos_tot = [dos[:, 1], dos[:, 2]]
        else:
            self.pdos_tot = [dos[:, 1]]
        self.pdos_energies = dos[:, 0] - self.efermi
        npoints = len(self.pdos_energies)
        # read in projections onto atomic orbitals
        self.pdos_atoms = {}
        self.pdos_kinds = {}
        for kind, info in self.pdos_info.items():
            pdos_kind = {}
            for istate, l in info['istate'].items():
                ncomponents = (2*l[0]+2) * self.nspins
                channel = '{0}{1}'.format(istate, orbitals[l[0]])
                pdos_kind[channel] = np.zeros((ncomponents, npoints), np.float)
            for iatom in info['iatom']:
                pdos_atom = {}
                for istate, l in info['istate'].items():
                    if self.spin_orbit_coupling:
                        filename = self.directory + '/projwfc/{0}.pdos_atm#{1}({2})_wfc#{3}({4}{5})'.format(
                            self.prefix, iatom, kind, istate, orbitals[l[0]], l[1])
                    else:
                        filename = self.directory + '/projwfc/{0}.pdos_atm#{1}({2})_wfc#{3}({4})'.format(
                            self.prefix, iatom, kind, istate, orbitals[l[0]])
                    channel = '{0}{1}'.format(istate, orbitals[l[0]])
                    pdosinp = np.genfromtxt(filename)
                    ncomponents = (2*l[0]+2) * self.nspins
                    pdos_atom[channel] = np.zeros(
                        (ncomponents, npoints), np.float)
                    for j in range(ncomponents):
                        pdos_atom[channel][j] += pdosinp[:, j + 1]
                        # sum over kind
                self.pdos_atoms[iatom] = pdos_atom
            self.pdos_kinds[kind] = pdos_kind
        self.pdos_kinds = self.merge_kinds(index=range(len(self.atoms)))
        # sum over orbital
        return self.pdos_energies, self.pdos_tot, self.pdos_atoms, self.pdos_kinds

    def merge_kinds(self, index=[]):
        '''
        merge atoms in the index
        '''
        pdos_kinds = {}
        npoints = len(self.pdos_energies)
        for kind, info in self.pdos_info.items():
            # print(kind, info)
            pdos_kind = self.pdos_kinds[kind]
            for channel, pdos in pdos_kind.items():
                # print(channel)
                pdos[:, :] = 0
            for iatom in info['iatom']:
                if iatom - 1 not in index:
                    continue
                # print(iatom)
                for channel, pdos_atom in self.pdos_atoms[iatom].items():
                    pdos_kind[channel] += pdos_atom
            pdos_kinds[kind] = pdos_kind
        return pdos_kinds

    def read_proj(self, ):
        '''
        read projwfc output file *.up and *.dw
        '''
        #
        self.read_pdos_info()
        proj_files = [self.directory + '/projwfc/%s.projwfc_up' %
                      self.prefix, self.directory + '/projwfc/%s.projwfc_down' % self.prefix]
        projs = []
        projs_kinds = []
        channels = []
        for i in range(self.nspins):
            with open(proj_files[i]) as f:
                lines = f.readlines()
                for j in range(len(lines)):
                    if '       1       1' in lines[j]:
                        break
                nstates = int(lines[j - 3].split()[0])
                nkpts = int(lines[j - 3].split()[1])
                nbands = int(lines[j - 3].split()[2])
                print(nstates, nkpts, nbands)
                proj = np.zeros((nstates, nkpts, nbands), np.float)
                proj_kinds = {}
                for istate in range(nstates):
                    iline = j + istate*(nkpts*nbands + 1) - 1
                    kinds = lines[iline].split()[2]
                    orbital = lines[iline].split()[3]
                    channel = kinds + orbital
                    if channel not in channels:
                        channels.append(channel)
                        proj_kinds[channel] = np.zeros(
                            (nkpts, nbands), np.float)
                    # l = lines[iline].split()[2]
                    # print(kinds, istate)
                    for ikpt in range(nkpts):
                        for iband in range(nbands):
                            iline = j + istate * \
                                (nkpts*nbands + 1) + ikpt*nbands + iband
                            # print('line: ', iline)
                            proj[istate, ikpt, iband] = float(
                                lines[iline].split()[2])
                            proj_kinds[channel][ikpt,
                                                iband] += float(lines[iline].split()[2])
                            # proj_atoms[channel, ikpt, iband] +=
            projs.append(proj)
            projs_kinds.append(proj_kinds)
        self.projs = projs
        self.projs_kinds = projs_kinds

    def plot_data(self, energies, dos, label, xindex, ax=None, fill=True, color=None, smearing=None):
        if ax is None:
            ax = plt.gca()
        for i in range(self.nspins):
            if not smearing:
                xmesh = energies[xindex]
                ymesh = dos[i][xindex]
            else:
                xmesh, ymesh = self.smearing(
                    energies[xindex], dos[i][xindex], sigma=smearing[0], de=smearing[0])
            ymesh = ymesh.reshape(-1, )
            if self.nspins == 2:
                newlabel = '%s-%s' % (label, lspins[i])
            else:
                newlabel = label
            if color is not None:
                ax.plot(xmesh, (-1)**i*ymesh, linewidth=0.7,
                        label=newlabel, color=color)
            else:
                ax.plot(xmesh, (-1)**i*ymesh, linewidth=0.7, label=newlabel)
            if fill:
                if color is not None:
                    ax.fill_between(xmesh, (-1)**i*ymesh, 0,
                                    alpha=0.2, color=color)
                else:
                    ax.fill_between(xmesh, (-1)**i*ymesh, 0, alpha=0.2,)
        return ax

    def plot_dos(self, energies=None, dos=None,
                 Emin=-5, Emax=5, color=None,
                 ax=None, fill=True, output=None, smearing=None):
        '''
        '''
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(figsize=(6, 3))
            # ax = plt.gca()
        if dos is None:
            dos = self.dos
        if energies is None:
            energies = self.dos_energies
        xindex = (energies > Emin) & (energies < Emax)
        self.plot_data(energies, dos, label='dos', ax=ax,
                       xindex=xindex, fill=fill, color=color, smearing=smearing)
        # ax.legend()
        # plt.grid(linestyle = '--')
        # ax.set_xlabel('Energy (eV)')
        # ax.set_ylabel('DOS (a.u.)')
        # ax.set_title('%s' % self.prefix)
        if output is not None:
            plt.savefig('%s' % output)
        return ax

    def plot_pdos_tot(self, energies=None, dos=None,
                      Emin=-5, Emax=5,
                      ax=None, fill=True, output=None, smearing=None):
        # print(self.pdos_tot)
        if dos is None:
            dos = self.pdos_tot
        if energies is None:
            energies = self.pdos_energies
        xindex = (energies > Emin) & (energies < Emax)
        fig, ax = plt.subplots(figsize=(6, 3))
        self.plot_data(energies, dos, label='pdos', ax=ax,
                       xindex=xindex, fill=fill, smearing=smearing)
        ax.legend()
        ax.set_xlabel('Energy (eV)')
        ax.set_ylabel('PDOS (a.u.)')
        ax.set_title('%s' % self.prefix)
        if output is not None:
            plt.savefig('%s' % output)
        return ax

    def plot_pdos(self, energies=None, pdos_kinds=None,
                  Emin=-5, Emax=5,
                  ax=None, total=False, select=None, fill=True,
                  output=None, legend=False, xylabel=True, smearing=None):
        '''
        '''
        if energies is None:
            energies = self.pdos_energies
        if pdos_kinds is None:
            pdos_kinds = self.pdos_kinds
        xindex = (energies > Emin) & (energies < Emax)
        # print(xindex)
        if ax is None:
            fig, ax = plt.subplots(figsize=(6, 3))
        # if total:
            # self.plot_data(self.pdos_energies, self.pdos_tot, label = 'pdos', ax = ax, fill = fill)
        for kind, channels in pdos_kinds.items():
            if select and kind not in select:
                continue
            number = chemical_symbols.index(kind)
            color = jmol_colors[number]
            for channel, pdos in channels.items():
                if select and channel[-1] not in select[kind]:
                    continue
                label = '{0}-{1}'.format(kind, channel)
                self.plot_data(energies, pdos, label=label, ax=ax,
                               xindex=xindex, fill=fill, color=color, smearing=smearing)
        if legend:
            ax.legend()
        if xylabel:
            ax.set_xlabel('Energy (eV)')
            ax.set_ylabel('PDOS (a.u.)')
        if output is not None:
            plt.savefig('%s' % output)
        return ax

    def plot_pdos_layer(self, atoms=None,
                        layers=None,
                        miller=(0, 0, 1),
                        dz=0.3,
                        Emin=-5,
                        Emax=3,
                        total=False,
                        select=None,
                        fill=True,
                        smearing=None,
                        output=None):
        '''

        '''
        atoms = self.atoms
        if not layers:
            layers = get_layers(atoms, miller, dz)[0]
        nlayers = max(layers) + 1
        fig, axs = plt.subplots(
            nlayers, 1, figsize=(10, nlayers*2), sharex=True)
        indexs = range(len(atoms))
        images = []
        xindex = (self.pdos_energies > Emin) & (self.pdos_energies < Emax)
        for ilayer in range(nlayers):
            iax = nlayers - ilayer - 1
            # axs[iax].set_yticks([])
            index = [j for j in indexs if layers[j] == ilayer]
            images.append(atoms[index])
            pdos_kinds = self.merge_kinds(index)
            self.plot_pdos(energies=self.pdos_energies, pdos_kinds=pdos_kinds, select=select,
                           Emin=Emin, Emax=Emax, ax=axs[iax],
                           legend=False, xylabel=False, fill=fill, smearing=smearing)
            axs[iax].axvline(0, color='b')
            # print(iax, ilayer)
            # print(index)
            # axs[iax].set_xlim([Emin, Emax])
        plt.xlabel('E - E$_{Fermi}$ (eV)', size='16')
        plt.subplots_adjust(hspace=0)
        fig.text(0.05, 0.5, 'PDOS (a.u.)', size='16',
                 va='center', rotation='vertical')
        if output is not None:
            # output = '{0}-pdos.png'.format(self.prefix)
            plt.savefig('%s' % output)
        # view(images)
        return axs, images

    def get_pdos(self, species, orbital):
        dos = self.pdos_kinds[species][orbital]
        return self.pdos_energies, dos

    def get_d_band_center(self, species=None, orbital='d'):
        '''
        '''
        energies, dos = self.get_pdos(species, orbital)
        Nstates = np.trapz(dos, energies)
        occupied = energies <= 0.0
        N_occupied_states = np.trapz(dos[occupied], energies[occupied])
        # first moment
        d_center = np.trapz(energies * dos, energies) / Nstates
        # second moment
        d_width = np.trapz(energies**2 * dos, energies) / Nstates
        return d_center, d_width

    def smearing(self, energies, dos, sigma=0.1, de=0.01, total=False):
        '''
        'natoms':         the total number of atoms of this kind in the structure
        'sigma':          sigma for the gaussian distribution (default: 0.01)
        'de':     integration step size (default: 0.01)
        '''
        npnts = len(energies)
        emin = min(energies)
        emax = max(energies)
        nmesh = int((emax-emin)/de)+1
        xmesh = np.linspace(emin, emax, nmesh)
        ymesh = np.zeros((nmesh, 1))
        fact = de/(sigma*np.sqrt(2.0*np.pi))
        for i in range(nmesh):
            func = np.exp(-(xmesh[i]-energies)**2/(2.0*sigma**2))*fact
            ymesh[i] = func.dot(dos[:])
        # ymesh /= natoms  # normalize
        # return energies, dos
        return xmesh, ymesh


def compare_pdos(dos1, index1, dos2, index2):
    '''
    '''
    if len(index1) != len(index2):
        assert ('number of atoms wrong!')
    natoms = len(index1)
    for i in range(natoms):
        if dos1.atoms[i].symbol != dos2.atoms[i].symbol:
            assert('atoms symbol not match!')
    ddos = copy.deepcopy(dos1)
    # print(dos1.atoms.get_chemical_symbols())
    # print(dos2.atoms.get_chemical_symbols())
    atoms = ddos.atoms
    for i in range(natoms):
        iatom1 = index1[i]
        iatom2 = index2[i]
        for channel in ddos.pdos_atoms[iatom1 + 1]:
            # print(i, atoms[i].symbol, channel)
            npoints = len(ddos.pdos_atoms[iatom1 + 1][channel][0])
            ddos.pdos_atoms[iatom1 + 1][channel] = dos1.pdos_atoms[iatom1 + 1][channel][:,
                                                                                        0:npoints] - dos2.pdos_atoms[iatom2 + 1][channel][:, 0:npoints]
    ddos.pdos_kinds = ddos.merge_kinds(index=range(len(ddos.atoms)))
    return ddos



from xespresso.post.base import PostCalculation


class EspressoQ2r(PostCalculation):

    package = 'q2r'
    package_parameters = {'INPUT': ['fildyn', 'flfrc', 'zasr', 'loto_2d'], }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)



from operator import pos
import xml.etree.ElementTree as ET
from pprint import pprint
from ase import Atoms, symbols
from ase.atom import Atom
from ase.atoms import default
from ase.units import create_units
from ase.utils import iofunction
import numpy as np
from xespresso.utils import modify_text

# Quantum ESPRESSO uses CODATA 2006 internally
units = create_units('2006')

different_names = {
    'CONTROL': {'forces': 'tprnfor', 'stress': 'tstress'}
}


def xml_parser(xmlfile):
    """
    """
    pwi_parameters = {}
    # section_names = ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL', 'ATOMIC_SPECIES', 'K_POINTS', 'CELL_PARAMETERS', 'CONSTRAINTS', 'OCCUPATIONS', 'ATOMIC_VELECITIES', 'ATOMIC_FORCES']
    section_names = ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL', ]
    for section_name in section_names:
        pwi_parameters[section_name] = {}
    #
    tree = ET.parse(xmlfile)
    root = tree.getroot()
    input = root.find('./input')
    parameters = {}
    subs = ['control_variables', 'spin', 'bands', 'basis',
            'electron_control', 'ion_control', 'cell_control']
    for sub in subs:
        parameters[sub] = {}
        for key in input.find('./%s' % sub):
            parameters[sub][key.tag] = key.text
            parameters[sub].update(key.attrib)
    #
    pwi_parameters['CONTROL'].update(parameters['control_variables'])
    pwi_parameters['SYSTEM'].update(parameters['bands'])
    pwi_parameters['SYSTEM'].update(parameters['basis'])
    pwi_parameters['ELECTRONS'].update(parameters['electron_control'])
    #
    for section_name, paras in different_names.items():
        for key, value in paras.items():
            pwi_parameters[section_name][value] = pwi_parameters[section_name].pop(
                key)

    # atomic_species
    atomic_species, pseudopotentials, input_ntyp = get_atomic_species(
        input.find('./atomic_species'))
    atoms = get_atomic_structure(input.find('./atomic_structure'))
    kpts, koffset = get_kpoint_grid(input.find('./k_points_IBZ'))
    #
    # dftU
    dftU_parameters = get_dftU(input.find('./dft'), input_ntyp)
    pwi_parameters['SYSTEM'].update(dftU_parameters)
    #
    if parameters['spin']['lsda'].upper() == 'TRUE':
        pwi_parameters['SYSTEM'].update({'nspin': 2})
    #
    pwi_parameters = xml2pw(pwi_parameters)
    # Ha to Ry
    for key in ['forc_conv_thr', 'etot_conv_thr']:
        if key in pwi_parameters['CONTROL']:
            pwi_parameters['CONTROL'][key] = pwi_parameters['CONTROL'][key]*2
    for key in ['ecutrho', 'ecutwfc', 'ecutfock']:
        if key in pwi_parameters['SYSTEM']:
            pwi_parameters['SYSTEM'][key] = pwi_parameters['SYSTEM'][key]*2
    #
    pwi_parameters['INPUT_NTYP'] = input_ntyp
    parameters = {
        'input_data': pwi_parameters,
        'pseudopotentials': pseudopotentials,
        'kpts': kpts,
        'koffset': koffset,
        # 'atomic_species': atomic_species,
    }
    return atoms, parameters


def get_atomic_structure(xml):
    """
    Parse atom.

    Parameters
    """
    nat = xml.attrib['nat']
    alat = xml.attrib['alat']
    atoms = Atoms()
    positions = []
    symbols = []
    all_species = []
    for item in xml.find('./atomic_positions'):
        species = item.attrib['name']
        all_species.append(species)
        symbol = species.split('_')[0].split('-')[0]
        if len(symbol) == 3:
            symbol = symbol[0:2]
        symbols.append(symbol)
        index = item.attrib['index']
        data = item.text.split()
        # These can be fractions and other expressions
        position = [float(data[0])*units['Bohr'], float(data[1])
                    * units['Bohr'], float(data[2])*units['Bohr']]
        positions.append(position)
    positions = np.array(positions)
    # positions = positions**units['Bohr']
    atoms = Atoms(symbols=symbols, positions=positions)
    atoms.arrays['species'] = all_species
    # cell
    cell = []
    for item in xml.find('./cell'):
        data = item.text.split()
        cell.append([float(data[0]), float(data[1]), float(data[2])])
    cell = np.array(cell)
    cell = cell*units['Bohr']
    atoms.cell = cell
    atoms.pbc = True
    return atoms


def get_atomic_species(xml):
    """
    Parameters
    """
    default = ['mass', 'pseudo_file']
    atomic_species = []
    pseudopotentials = {}
    input_ntyp = {}
    for sub in xml.findall('./species'):
        species = {}
        name = sub.attrib['name']
        species['name'] = name
        for item in sub:
            species[item.tag] = item.text
            if item.tag not in default:
                if item.tag not in input_ntyp:
                    input_ntyp[item.tag] = {}
                input_ntyp[item.tag][name] = float(item.text)
        pseudopotentials[name] = species['pseudo_file']
        atomic_species.append(species)
    return atomic_species, pseudopotentials, input_ntyp


def get_dftU(xml, input_ntyp):
    """
    Parameters
    """
    dftU_parameters = {}
    if not xml.find('./dftU'):
        return dftU_parameters

    for item in xml.find('./dftU'):
        if item.tag == 'Hubbard_U':
            dftU_parameters['lda_plus_u'] = True
            if 'Hubbard_U' not in input_ntyp:
                input_ntyp['Hubbard_U'] = {}
            input_ntyp['Hubbard_U'][item.attrib['specie']] = float(
                item.text)/0.073498810939358  # Ry to eV
        else:
            dftU_parameters[item.tag] = item.text
    # print(input_ntyp)
    return dftU_parameters


def get_kpoint_grid(xml):
    """
    Parameters
    """
    data = xml[0].attrib
    kpts = (int(data['nk1']), int(data['nk2']), int(data['nk3']))
    koffset = (int(data['k1']), int(data['k2']), int(data['k3']))
    return kpts, koffset


def xml2pw(parameters, package='PW'):
    import copy
    from xespresso.input_parameters import qe_namespace
    new_parameters = copy.deepcopy(parameters)
    for section, paras in parameters.items():
        for key, value in paras.items():
            if key not in qe_namespace[package][section]:
                del new_parameters[section][key]
            else:
                # print(key, parameters[section][key], qe_namespace[package][section][key][0])
                # print(key, parameters[section][key], qe_namespace[package][section][key][1])
                new_parameters[section][key] = modify_text(
                    parameters[section][key], type=qe_namespace[package][section][key][0])
                # if not compare_value(new_parameters[section][key], qe_namespace[package][section][key][1], qe_namespace[package][section][key][0]):
                # del new_parameters[section][key]
    return new_parameters


if __name__ == "__main__":
    xmlfile = '/home/xing/ase/xespresso/restart/scf/fe+u/fe+u.xml'
    xmlfile = '/home/xing/ase/xespresso/restart/relax/h2o/h2o.xml'
    xml_data = xml_parser(xmlfile)
    print(xml_data['atoms'])
    print('pseudopotentials: ', xml_data['pseudopotentials'])
    pprint(xml_data['parameters'])



import numpy as np
from xespresso import Espresso
import matplotlib.pyplot as plt
import os
import shutil
import subprocess


class COHP:
    def __init__(self,
                 directory='.',
                 prefix='cohp',
                 indexs=[[1, 2]],
                 queue=False,
                 command='lobster',
                 **kwargs):
        """
        Energetic window [COHPstartEnergy, COHPendEnergy]

        Args:
            command (str, optional): Path to lobster executable file.
        """
        self.directory = directory
        self.prefix = prefix
        self.parameters = kwargs
        self.indexs = indexs
        self.command = command

    def run(self, cpu=1):
        """Run COHP analysis using lobster

        Args:
            cpu (int, optional): CPU threads used. Defaults to 1.

        Raises:
            EnvironmentError: [description]
        """
        shutil.copy(os.path.join(self.directory, '%s.pwi' % self.prefix),
                    os.path.join(self.directory, '%s.scf.in' % self.prefix))
        self.write_input()
        command = self.command
        os.system('export OMP_NUM_THREADS=%s' % cpu)
        print('Running %s on %s cpus' % (command, cpu))
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Command "{}" failed in '
                   '{} with error code {}'.format(command, path, errorcode))
            print(msg)
            exit()
        print('Done: %s' % command)
        #

    def write_input(self, ):
        """make input file for lobster
        """
        filename = os.path.join(self.directory, 'lobsterin')
        with open(filename, 'w') as f:
            for key, value in self.parameters.items():
                if isinstance(value, list):
                    for subvalue in value:
                        f.write('  %s  %s \n' % (key, subvalue))
                else:
                    f.write('  %s  %s \n' % (key, value))
            for ind in self.indexs:
                f.write('cohpbetween atom %s atom %s \n' % (ind[0], ind[1]))

    def read_cohp(self, ):
        """read COHP anlysis results
        """
        from ase.calculators.vasp import VaspDos
        dos = VaspDos(os.path.join(self.directory, 'DOSCAR.lobster'))
        self.dos = dos._total_dos[1]
        self.dos_energies = dos._total_dos[0]
        # read COHP results
        datas = np.genfromtxt(os.path.join(self.directory, 'COHPCAR.lobster'),
                              skip_header=3 + len(self.indexs))
        self.cohp = datas[:, 1]
        self.cohp_energies = datas[:, 0]
        # read COOP results
        datas = np.genfromtxt(os.path.join(self.directory, 'COOPCAR.lobster'),
                              skip_header=3 + len(self.indexs))
        self.coop = datas[:, 1]
        self.coop_energies = datas[:, 0]
        # read COBI results
        datas = np.genfromtxt(os.path.join(self.directory, 'COBICAR.lobster'),
                              skip_header=3 + len(self.indexs))
        self.cobi = datas[:, 1]
        self.cobi_energies = datas[:, 0]

    def read_icohp(self, ):
        """read icohp data for bond strength evaluation
        """
        from ase.calculators.vasp import VaspDos
        import pandas as pd
        # read icohp
        datas = np.genfromtxt(os.path.join(self.directory,
                                           'ICOHPLIST.lobster'),
                              dtype=[int, 'S5', 'S5', float, float],
                              usecols=(0, 1, 2, 3, 7),
                              names=['interaction', 'atom1',
                                     'atom2', 'distance', 'ICOHP'],
                              skip_header=1)
        self.icohp = pd.DataFrame(datas)
        # self.icohp_energies = datas[:, 7]
        # read icoop
        datas = np.genfromtxt(os.path.join(self.directory,
                                           'ICOOPLIST.lobster'),
                              dtype=[int, 'S5', 'S5', float, float],
                              usecols=(0, 1, 2, 3, 7),
                              names=['interaction', 'atom1',
                                     'atom2', 'distance', 'ICOOP'],
                              skip_header=1)
        self.icoop = pd.DataFrame(datas)
        # self.icoop_energies = datas[:, 7]
        # read icobi
        datas = np.genfromtxt(os.path.join(self.directory,
                                           'ICOHPLIST.lobster'),
                              dtype=[int, 'S5', 'S5', float, float],
                              usecols=(0, 1, 2, 3, 7),
                              names=['interaction', 'atom1',
                                     'atom2', 'distance', 'ICOBI'],
                              skip_header=1)
        self.icobi = pd.DataFrame(datas)
        # self.icobi_energies = datas[:, 7]

    def plot_cohp(self, ax=None, output=None):
        """create a DOS and COHP plot

        Args:
            ax (COHP object, optional):  Defaults to None.
            output (strng, optional): filename if want to save plot. Defaults to None.

        Returns:
            plot: a DOS with COHP plot
        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1, 2)
        ax[0].plot(self.dos, self.dos_energies)
        ax[0].set_ylabel('E-Ef (eV)')
        ax[0].axhline(y=0, color='k', linewidth=0.5, linestyle='--')
        ax[1].plot(-self.cohp, self.cohp_energies)
        ax[1].axvline(x=0, color='k', linewidth=0.2)
        ax[1].axhline(y=0, color='k', linewidth=0.2, linestyle='--')
        ax[0].set_xlabel('DOS')
        ax[1].set_xlabel('-pCOHP')
        if output:
            plt.savefig(output)
        return ax

    def plot_coop(self, ax=None, output=None):
        """create a DOS with COOP plot

        Args:
            ax (COHP object, optional): Defaults to None.
            output (string, optional): filename if want to save plot. Defaults to None.

        Returns:
            plot: a DOS with COOP plot
        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1, 2)
        ax[0].plot(self.dos, self.dos_energies)
        ax[0].set_ylabel('E-Ef (eV)')
        ax[0].axhline(y=0, color='k', linewidth=0.5, linestyle='--')
        ax[1].plot(self.coop, self.coop_energies)
        ax[1].axvline(x=0, color='k', linewidth=0.5)
        ax[1].axhline(y=0, color='k', linewidth=0.5, linestyle='--')
        ax[0].set_xlabel('DOS')
        ax[1].set_xlabel('pCOOP')
        if output:
            plt.savefig(output)
        return ax

    def plot_cobi(self, ax=None, output=None):
        """create a DOS with COBI plot

        Args:
            ax (COHP object, optional): Defaults to None.
            output (string, optional): filename if want to save plot. Defaults to None.

        Returns:
            plot: a DOS with COBI plot
        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1, 2)
        ax[0].plot(self.dos, self.dos_energies)
        ax[0].set_ylabel('E-Ef (eV)')
        ax[0].axhline(y=0, color='k', linewidth=0.5, linestyle='--')
        ax[1].plot(self.cobi, self.cobi_energies)
        ax[1].axvline(x=0, color='k', linewidth=0.5)
        ax[1].axhline(y=0, color='k', linewidth=0.5, linestyle='--')
        ax[0].set_xlabel('DOS')
        ax[1].set_xlabel('COBI')
        if output:
            plt.savefig(output)
        return ax




from xespresso.post.base import PostCalculation


class EspressoProjwfc(PostCalculation):

    package = 'projwfc'
    package_parameters = {'PROJWFC': ['prefix', 'outdir', 'ngauss', 'degauss',
                                      'Emin', 'Emax', 'DeltaE', 'lsym', 'pawproj', 'filpdos', 'filproj',
                                      'lwrite_overlaps', 'lbinary_data', 'kresolveddos', 'tdosinboxes',
                                      'n_proj_boxes', 'irmin(3,n_proj_boxes)', 'irmax(3,n_proj_boxes)',
                                      'plotboxes', ]
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)




from xespresso.post.base import PostCalculation


class EspressoPh(PostCalculation):

    package = 'ph'
    package_parameters = {'INPUTPH': ['amass', 'outdir', 'prefix', 'niter_ph', 'tr2_ph',
                                      'alpha_mix(niter)', 'nmix_ph', 'verbosity', 'reduce_io',
                                      'max_seconds', 'fildyn', 'fildrho', 'fildvscf', 'epsil',
                                      'lrpa', 'lnoloc', 'trans', 'lraman', 'eth_rps', 'eth_ns',
                                      'dek', 'recover', 'low_directory_check', 'only_init', 'qplot',
                                      'q2d', 'q_in_band_form', 'electron_phonon', 'el_ph_nsigma',
                                      'el_ph_sigma', 'ahc_dir', 'ahc_nbnd', 'ahc_nbndskip',
                                      'skip_upperfan', 'lshift_q', 'zeu', 'zue', 'elop', 'fpol',
                                      'ldisp', 'nogg', 'asr', 'ldiag', 'lqdir', 'search_sym',
                                      'nq1', 'nq2', 'nq3', 'nk1', 'nk2', 'nk3', 'k1', 'k2', 'k3',
                                      'diagonalization', 'read_dns_bare', 'ldvscf_interpolate',
                                      'wpot_dir', 'do_long_range', 'do_charge_neutral', 'start_irr',
                                      'last_irr', 'nat_todo', 'modenum', 'start_q', 'last_q',
                                      'dvscf_star', 'drho_star', ],
                          'LINE': ['xq', 'atom'],
                          # 'qPointsSpecs': ['nqs', 'xq1', 'xq2', 'xq3', 'nq'],
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)




from ase.calculators.calculator import FileIOCalculator, CalculationFailed, equal, compare_atoms, PropertyNotPresent
from xespresso.xio import read_espresso_asei
from xespresso.xespresso import Espresso
import os
import logging
logger = logging.getLogger(__name__)


class EspressoNscf(FileIOCalculator):

    package = 'pw'

    def __init__(self, scf_directory, prefix, parallel='',
                 queue=None, debug=False,
                 kpts=(10, 10, 10),
                 **kwargs):
        print("{0:=^60}".format('nscf'))
        self.scf_directory = scf_directory
        self.prefix = prefix
        self.parallel = parallel
        self.queue = queue
        if debug:
            logger.setLevel(debug)
        self.load_scf()
        self.parameters['kpts'] = kpts
        self.parameters.update(kwargs)
        self.parameters = Espresso.check_input(self.parameters, self.prefix)
        #
        self.directory = os.path.join(self.scf_directory, 'nscf')
        self.label = os.path.join(self.directory, self.prefix)
        self.asei = os.path.join(self.directory, '%s.nscf_asei' % self.prefix)
        # create working directory
        self.set_label(self.directory, self.prefix)

    def set_label(self, label, prefix):
        '''
        set directory and prefix from label
        '''
        self.directory = label
        if not prefix:
            self.prefix = os.path.split(label)[1]
        else:
            self.prefix = prefix
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        self.label = os.path.join(self.directory, self.prefix)
        self.pwi = os.path.join(self.directory, '%s.pwi' % self.prefix)
        self.pwo = os.path.join(self.directory, '%s.pwo' % self.prefix)
        self.save_directory = os.path.join(
            self.directory, '%s.save' % self.prefix)
        logger.debug('Directory: %s' % (self.directory))
        logger.debug('Prefix: %s' % (self.prefix))

    def load_scf(self):
        """load parameters and results from privous scf calculation
        """
        import copy
        asei = os.path.join(self.scf_directory, '%s.asei' % self.prefix)
        self.atoms, self.scf_parameters = read_espresso_asei(asei, 'PW')
        self.parameters = copy.deepcopy(self.scf_parameters)
        self.parameters['input_data']['CONTROL']['calculation'] = 'nscf'
        self.parameters['input_data']['CONTROL']['verbosity'] = 'high'
        self.parameters['input_data']['CONTROL']['outdir'] = '../'
        self.parameters['input_data']['SYSTEM']['occupations'] = 'tetrahedra'
        self.parameters['input_data']['SYSTEM'].pop('degauss', None)

    def write_input(self, atoms, properties=None, system_changes=None):
        from xespresso.xio import write_espresso_asei, write_espresso_in
        from xespresso.scheduler import set_queue
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        set_queue(self, package=self.package, parallel=self.parallel,
                  queue=self.queue)
        write_espresso_in(self.label + '.pwi', atoms, **self.parameters)
        logger.debug("write asei file: {}".format(self.asei))
        logger.debug("  with parameters: {}".format(self.parameters))
        write_espresso_asei(self.asei, self.state_info, self.parameters)
        logger.debug('Write input successfully')

    @property
    def state_info(self):
        from xespresso.utils import get_hash
        chargeFile = os.path.join(
            self.scf_directory, '%s.save/charge-density.dat' % self.prefix)
        state_info = get_hash(chargeFile)
        return state_info

    def check_state(self, ):
        # read information of the charge-density file
        self.state_parameters = self.parameters
        output, meg = self.read_convergence_post('pw')
        logger.debug("Check state: {}".format(meg))
        if output:
            if os.path.isfile(self.asei):
                system_changes = self.check_state_post(self.asei, package='PW')
                if not system_changes:
                    logger.debug('Use previous results!')
                    return False
            else:
                logger.debug('No asei output. Start a new calculation.')

        else:
            logger.debug('No pw output. Start a new calculation.')

        return True

    def check_state_post(self, asei, package):
        logger.debug("asei file: {}".format(asei))
        old_state_info, old_state_parameters = read_espresso_asei(
            asei, package)
        if not self.state_info == old_state_info:
            logger.debug('File in save changed')
            return True
        elif not self.state_parameters == old_state_parameters:
            logger.debug(self.state_parameters)
            logger.debug(old_state_parameters)
            logger.debug('Parameters changed')
            return True
        else:
            return False

    def read_convergence_post(self, package='pw'):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        }
        '''
        output = self.label + '.%so' % package
        logger.debug("read output: {}".format(output))
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return False, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0:
                return False, 'pwo file has nothing'
            nlines = len(lines)
            n = min([100, nlines])
            for line in lines[-n:-1]:
                if line.rfind('JOB DONE.') > -1:
                    logger.debug('JOB DONE.')
                    return True, line
        return False, line

    def run(self):
        if self.check_state():
            self.calculate()



import os
import numpy as np
from copy import deepcopy
from ase.atoms import Atoms
from xespresso import Espresso
from xespresso.workflow.base import Base
from xespresso.xlog import XLogger
from phonopy import Phonopy
import numpy as np

import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
import subprocess


class PLogger(XLogger):
    """Class for handling all text output."""

    def __init__(self, ):
        XLogger.__init__(self,)

    def logo(self):
        self()
        self(' //===\\  ||   ||   //==\\  ')
        self(' ||   ||  ||   ||  ||    || ')
        self(' ||===//  ||===||  ||    || ')
        self(' ||       ||   ||  ||    ||  ')
        self(' ||       ||   ||   \\==//   ')
        self()


# view(mols.values())
class Phonon(Base):
    def __init__(self, atoms, label='.', prefix=None, calculator=None, view=False):
        Base.__init__(self, atoms, label=label, prefix=prefix,
                      calculator=calculator, view=view)
        self.set_logger(PLogger)
        self.supercells = {}
        self.set_of_forces = []

    def run(self):
        pass

    def phonopy(self,
                supercell_matrix=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                primitive_matrix=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                distance=0.01):
        '''
        '''
        self.build_phonon(supercell_matrix=supercell_matrix,
                          primitive_matrix=primitive_matrix,
                          distance=distance)
        self.build_supercells()
        self.pool_atoms(self.supercells)
        self.post()

    def build_phonon(self, atoms=None,
                     supercell_matrix=None,
                     primitive_matrix=None,
                     distance=None):
        if atoms is None:
            atoms = self.atoms
        phonon = Phonopy(atoms,
                         supercell_matrix,
                         primitive_matrix=primitive_matrix)
        phonon.generate_displacements(distance=distance)
        self.log("[Phonopy] Atomic displacements:")
        disps = phonon.get_displacements()
        for d in disps:
            self.log("[Phonopy] %d %s" % (d[0], d[1:]))
        self.phonon = phonon

    def build_supercells(self):
        supercells = self.phonon.get_supercells_with_displacements()
        # Force calculations by calculator
        i = 0
        for scell in supercells:
            atoms = Atoms(symbols=scell.get_chemical_symbols(),
                          scaled_positions=scell.get_scaled_positions(),
                          cell=scell.get_cell(),
                          pbc=True)
            self.supercells[str(i).zfill(3)] = atoms
            i += 1
        self.nsupercells = len(self.supercells)
        self.log('')
        self.log('Number of supercells: %s' % (i))

    def post(self):
        # read forces
        for i in range(self.nsupercells):
            job = str(i).zfill(3)
            forces = self.results[job]['forces']
            drift_force = forces.sum(axis=0)
            self.log(("[Phonopy] Drift force:" + "%11.5f" * 3) %
                     tuple(drift_force))
            # Simple translational invariance
            for force in forces:
                force -= drift_force / forces.shape[0]
            self.set_of_forces.append(forces)
        #
        self.phonon.produce_force_constants(forces=self.set_of_forces)
        self.log('')
        self.log("[Phonopy] Phonon frequencies at Gamma:")
        for i, freq in enumerate(self.phonon.get_frequencies((0, 0, 0))):
            self.log("[Phonopy] %3d: %10.5f THz" % (i + 1, freq))  # THz

        # DOS
        self.phonon.set_mesh([41, 41, 41])
        self.phonon.set_total_DOS(tetrahedron_method=True)
        self.log('')
        self.log("[Phonopy] Phonon DOS:")
        self.omega, self.phdos = self.phonon.get_total_DOS()
        self.phonon_energies = 0.00414*self.omega
        for omega, dos in np.array(self.phonon.get_total_DOS()).T:
            self.log("%15.7f%15.7f" % (omega, dos))

    def opt(self, **kwargs):
        # vc-relax
        self.log('')
        self.log('Optmizatiion:')
        self.old_calculator = self.calculator.copy()
        self.calculator.update({'calculation': 'vc-relax'})
        self.calculator.update(kwargs)
        self.pool_atoms({'vc-relax': self.atoms})
        self.atoms = self.results['vc-relax']['atoms']
        self.calculator = self.old_calculator
        # scf
        # self.calculator.update({'calculation': 'scf'})
        # self.pool_atoms({'scf': self.atoms})

    def run_atoms(self, job, atoms):
        self.log('-'*60)
        self.log('Submit job {0}'.format(job))
        self.log.print_atoms(atoms)
        calc = Espresso(
            label=os.path.join(self.label, job),
            **self.calculator,
        )
        calc.parameters['input_data']['prefix'] = job
        # self.log('    Commnad: {0}'.format(calc.command))
        atoms.calc = calc
        # atoms.get_potential_energy()
        calc.run(atoms=atoms, restart=1)
        calc.read_results()
        self.results[job] = deepcopy(calc.results)
        calc.clean()
        return job, self.results[job]['energy']

    def dfpt(self, job, queue):
        '''
        '''
        atoms = self.atoms
        self.log('-'*60)
        self.log('Submit job {0}'.format(job))
        self.log.print_atoms(atoms)
        calc = Espresso(
            label=os.path.join(self.label, job),
            **self.calculator,
        )
        calc.parameters['input_data']['prefix'] = job
        atoms.calc = calc
        energy = atoms.get_potential_energy()
        # dos
        calc.post(queue=queue,
                  package='ph',
                  tr2_ph=1e-14,
                  ldisp=True,
                  nq1=4,
                  nq2=4,
                  nq3=3,
                  )
        calc.post(queue=queue,
                  fildyn='matdyn',
                  package='q2r',
                  zasr='simple',
                  flfrc='%s.fc' % job,
                  )
        calc.post(queue=queue,
                  package='matdyn',
                  asr='simple',
                  dos=True,
                  flfrc='%s.fc' % job,
                  fldos='%s.phdos' % job,
                  nk1=40,
                  nk2=40,
                  nk3=30,
                  )
        calc.plot_phdos(fldos='%s' % job,
                        output='images/%s-phdos.png' % job)
        pass

    def read_phdos(self, fldos=None, ax=None, output=None):
        '''
        '''
        import matplotlib.pyplot as plt
        if fldos is None:
            fldos = self.prefix
        phdos = np.loadtxt(self.directory+'/%s.phdos' % fldos)
        self.phdos = phdos

    def plot_phdos(self, Emin=-5, Emax=20, ax=None, output=None):
        '''
        '''
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(figsize=(6, 3))
            # ax = plt.gca()
        xindex = (self.omega > Emin) & (self.omega < Emax)
        ax.plot(self.omega[xindex], self.phdos[xindex], linewidth=0.7)
        ax.set_xlabel('Frequency (THz)')
        ax.set_ylabel('Phonon DOS (a.u.)')
        plt.tight_layout()
        if output:
            plt.savefig('%s' % output)
        return ax



import os
import numpy as np
from copy import deepcopy
from time import sleep
import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
import multiprocessing
from xespresso import Espresso
from xespresso.tools import mypool, fix_layers, dipole_correction


class Base():
    def __init__(self, atoms, label='.', prefix=None, calculator=None, view=False, debug=False):
        self.set_label(label, prefix)
        self.atoms = atoms
        self.calculator = calculator
        self.view = view
        self.debug = debug
        self.images = {}
        self.children = {}
        self.results = {}
        # self('%s'%self.__class__.__name__)

    def set_logger(self, logger):
        self.log = logger()
        # self.log.fd = os.path.join(self.label, '%s.txt'%(self.prefix))
        self.log.fd = os.path.join(self.label, '%s.%so' %
                                   (self.prefix, self.name))
        self.log('-'*60)
        self.log('Class:  %s\n' % self.__class__.__name__)
        self.log('directory: %s' % self.directory)
        self.log('prefix: %s' % self.prefix)
        self.log('-'*60)
        self.log('Atomic structure:')
        self.log.print_atoms(self.atoms)
        # self.log('Calculator parameters:')
        # self.log.print_dict(self.calculator)

    def set_label(self, label, prefix):
        self.label = label
        if not prefix:
            self.directory = os.path.split(label)[0]
            self.prefix = os.path.split(label)[1]
        else:
            self.directory = label
            self.prefix = prefix
        if not os.path.exists(label):
            os.makedirs(label)

    def view_images(self, images=None):
        """
        """
        from ase.visualize import view
        if not images:
            images = self.images
        print('%s, Number of images: %s' % (self.prefix, len(images)))
        if len(images) == 0:
            return
        view(images.values())

    def run(self):
        self.build_children()
        self.run_children()

    def run_children(self):
        from random import random
        max_workers = len(self.children)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for job, child in self.children.items():
                sleep(random())
                futures.append(executor.submit(child.run))
            for future in as_completed(futures):
                print(future.result())

    def run_children_multiprocessing(self, showInfo=False):
        '''
        '''
        from random import random
        from time import sleep
        pool = multiprocessing.Pool(len(self.children))
        results = []
        for job, child in self.children.items():
            sleep(random())
            results.append(pool.apply_async(child.run))
        for r in results:
            r.get()
        pool.close()
        pool.join()

    def cell_relax_espresso(self, job, atoms):
        """
        """
        self.log('-'*60)
        self.log('Submit job {0}'.format(job))
        self.log.print_atoms(atoms)
        calculator = deepcopy(self.calculator)
        calculator.update({'calculation': 'vc-relax',
                           'prefix': job, })
        calc = Espresso(
            label=os.path.join(self.label, job),
            **calculator,
        )
        atoms.calc = calc
        calc.run(atoms=atoms)
        calc.read_results()
        self.results[job] = deepcopy(calc.results)
        return job, self.results[job]['energy']

    def geo_pdos_zpe(self, job, atoms):
        self.geo_relax_espresso(job, atoms)
        atoms = self.results[job]['atoms']
        self.vib_zpe(job, atoms)

    def get_kpts(self, atoms):
        """
        Generate k-point by density or unit cell.
        """
        a, b, c, alpha, beta, gamma = atoms.get_cell_lengths_and_angles()
        kpts = (int(30/a), int(30/b), int(30/c))
        return kpts

    def geo_relax_espresso(self, job, atoms):
        """
        """
        self.log('-'*60)
        self.log('Run geometry relaxtion: {0}'.format(job))
        self.log.print_atoms(atoms)
        calculator = deepcopy(self.calculator)
        calculator.update({'calculation': 'relax',
                           'prefix': job, })
        dip = dipole_correction(atoms, edir=3)
        calculator.update(dip)
        calc = Espresso(
            label=os.path.join(self.label, job),
            **calculator,
        )
        atoms.calc = calc
        calc.run(atoms=atoms)
        calc.read_results()
        energy = calc.results['energy']
        self.results[job] = deepcopy(calc.results)
        self.calcs[job] = calc
        return job, energy

    def pdos(self, job, atoms):
        """
        calculate the pdos
        """
        self.log('-'*60)
        self.log('Run pdos calculation: {0}'.format(job))
        calc = self.calcs[job]
        fe = calc.get_fermi_level()
        atoms = calc.results['atoms']
        calculator = deepcopy(self.calculator)
        kpts = [calculator['kpts'][0]*2, calculator['kpts']
                [1]*2, calculator['kpts'][2]*2]
        if 'parallel' not in calculator:
            calculator['parallel'] = '-nk 1'
        calc.nscf(queue=calculator['queue'], parallel=calculator['parallel'],
                  occupations='tetrahedra', kpts=kpts)
        calc.nscf_calculate()
        calc.read_results()
        calc.post(queue=calculator['queue'], package='dos',
                  Emin=fe - 30, Emax=fe + 30, DeltaE=0.01)
        calc.post(queue=calculator['queue'], package='projwfc',
                  Emin=fe - 30, Emax=fe + 30, DeltaE=0.01)

    def pool_atoms(self, jobs, func):
        from random import random
        if self.debug:
            for job, atoms in jobs.items():
                func(job, atoms)
            return 0
        max_workers = len(jobs)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for job, atoms in jobs.items():
                sleep(random())
                futures.append(executor.submit(func, job, atoms))
            for future in as_completed(futures):
                print(future.result())

    def build_children(self):
        pass

    def run_command(self, command=None):
        import subprocess
        from ase.calculators.calculator import CalculationFailed
        if command is None:
            command = self.command
        print('Running %s' % command)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err
        errorcode = proc.wait()
        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Command "{}" failed in '
                   '{} with error code {}'.format(command,
                                                  path, errorcode))
            raise CalculationFailed(msg)
        print('Done')



import os
from ase.geometry import get_layers
from ase.constraints import FixAtoms
from ase.io.espresso import construct_namelist
from xespresso import Espresso
from xespresso.xio import build_atomic_species_str
from ase.dft.bandgap import bandgap
import pickle
import multiprocessing
import numpy as np
# ====================================================


def get_nbnd(atoms=None, scale=1.2, pseudopotentials={}, nspin=1, input_data={}):
    input_parameters = construct_namelist(input_data)
    atomic_species_str, species_info, total_valence = build_atomic_species_str(
        atoms, input_parameters, pseudopotentials)
    nbnd = int(total_valence/(2.0/nspin))
    nbnd_scale = int(nbnd*scale)
    print(' total valence: %s\n nbnd: %s\n scaled nbnd: %s' %
          (total_valence, nbnd, nbnd_scale))
    return nbnd


def merge_slab(slab1, slab2, index=2):
    '''	
    '''
    slab2.cell[index] = slab1.cell[index]
    slab2.set_cell(slab1.cell, scale_atoms=True)
    slab1 = slab1 + slab2
    slab1.cell[2][2] = max(slab1.positions[:, 2] + 15)
    slab1.wrap()
    return slab1


def qeinp(calculation, ecutwfc=30, mixing_beta=0.5, conv_thr=1.0e-8,
          edir=False, input_ntyp={}, atoms=None):
    #
    inp = {
        # control
        'calculation':         calculation,
        'max_seconds':         78000,
        'verbosity':           'high',
        'tprnfor':             True,
        # system
        'ecutwfc':             ecutwfc,
        'ecutrho':             ecutwfc*8,
        'occupations':         'smearing',
        'degauss':             0.01,
        'input_ntyp':          input_ntyp,
        # electrons
        'mixing_beta':         mixing_beta,
        'conv_thr':            conv_thr,
        'electron_maxstep':    400,
    }
    #
    if edir:
        inp.update(dipole_correction(atoms, edir))

    return inp


def dipole_correction(atoms, edir=3):
    inp = {
        'dipfield':  True,
        'tefield':   True,
        'edir':      edir,
        'eamp':      0.001,
        'eopreg':    0.05,
    }
    emaxpos = max(atoms.positions[:, edir - 1] + atoms.cell[edir - 1]
                  [edir - 1])/2.0/atoms.cell[edir - 1][edir - 1]
    inp['emaxpos'] = round(emaxpos, 2)
    return inp

# tools


def build_oer(atoms):
    '''
    '''
    from ase.atoms import Atoms
    # ----------------------------------
    ooh = Atoms('O2H', positions=[[0, 0, 0], [1.4, 0, 0], [1.4, 0, 1.0]])
    ooh.rotate('z', np.pi/4)
    mols = {
        'o': Atoms('O'),
        'oh': Atoms('OH', positions=[[0, 0, 0], [0, 0, 1.0]]),
        'ooh': ooh,
    }
    #
    jobs = {}
    maxz = max(atoms.positions[:, 2])
    indm = [atom.index for atom in atoms if atom.z >
            maxz - 1.0 and atom.symbol != 'O'][0]
    for job, mol in mols.items():
        # print(job, mol)
        ads = mol.copy()
        natoms = atoms.copy()
        ads.translate(atoms[indm].position - ads[0].position + [0, 0, 1.9])
        natoms = natoms + ads
        jobs[job] = natoms
    return jobs


def fix_layers(atoms, miller=(0, 0, 1), tol=1.0, n=[0, 4]):
    '''
    '''
    layers = get_layers(atoms, miller, tol)[0]
    index = [j for j in range(len(atoms)) if layers[j] in range(n[0], n[1])]
    constraint = FixAtoms(indices=index)
    atoms.set_constraint(constraint)
    return atoms


def mypool(jobs, func, showInfo=False):
    '''
    '''
    from random import random
    from time import sleep
    pool = multiprocessing.Pool(len(jobs))
    results = []
    images = []
    for job, atoms in jobs.items():
        if showInfo:
            print(job, len(atoms), atoms)
        sleep(random()*2)
        results.append(pool.apply_async(func, (job, atoms)))
    for r in results:
        r.get()
    pool.close()
    pool.join()


def dwubelix(updates=[]):
    file = 'pw err out dos pdos projwfc int xyz path a.xml txt png'
    print('Downloading.....')
    cwd = os.getcwd()
    for update in updates:
        os.chdir(update)
        os.system('dwubelix-sc.py %s' % file)
        os.chdir(cwd)
    print('Finished')


def ana(dire, calc):
    atoms = calc.results['atoms']
    results = [dire, atoms, atoms.cell, atoms.positions]
    for prop in ['energy', 'forces', 'stress', 'magmoms']:
        if prop in calc.results:
            prop = calc.results[prop]
        else:
            prop = None
        results.append(prop)
    return results
# tools


def summary(updates=[], prefix='datas'):
    import pandas as pd
    columns = ['label', 'atoms', 'cell',
               'positions', 'energy', 'forces', 'stress']
    file = '%s.pickle' % prefix
    db = '%s.db' % prefix
    if os.path.exists(file):
        with open(file, 'rb') as f:
            datas, df = pickle.load(f)
    else:
        datas = {}
        df = pd.DataFrame(columns=columns)
    calc = Espresso()
    print('Reading.....')
    for update in updates:
        cwd = os.getcwd()
        for i, j, y in os.walk(update):
            output = is_espresso(i)
            if output:
                os.chdir(i)
                print('='*30)
                print('Reading dire:', i)
                calc.directory = cwd + '/' + i
                calc.prefix = output[0:-4]
                try:
                    calc.results = {}
                    calc.read_results()
                    datas[i] = calc.results
                    atoms = calc.results['atoms']
                    atoms.write(os.path.join(
                        calc.directory, '%s.cif' % calc.prefix))
                except Exception as e:
                    print('error: %s \n' % e)
            os.chdir(cwd)
    with open(file, 'wb') as f:
        pickle.dump([datas, df], f)
    print('Finished')


def is_espresso(path):
    '''
    check espresso 
    '''
    dirs = os.listdir(path)
    # print(dirs)
    # flag = True

    for qefile in ['.pwi']:
        flag = False
        for file in dirs:
            if qefile in file:
                return file
        if not flag:
            return False
    # return flag


def grep_valence_configuration(pseudopotential):
    """
    Given a UPF pseudopotential file, find the valence configuration.

    Valence configuration:
    nl pn  l   occ       Rcut    Rcut US       E pseu
    3S  1  0  2.00      0.700      1.200    -6.910117

    """
    orbitals = ['S', 'P', 'D', 'F']
    valence = {}
    with open(pseudopotential) as psfile:
        lines = psfile.readlines()
        for i in range(len(lines)):
            if 'valence configuration:' in lines[i].lower():
                j = i + 2
                ob = lines[j].split()[0]
                while ob[1] in orbitals:
                    valence[ob] = lines[j].split()[3]
                    j += 1
                    ob = lines[j].split()[0]
                return valence
    if not valence:
        raise ValueError(
            'Valence configuration missing in {}'.format(pseudopotential))



"""Quantum ESPRESSO Calculator

export PYTHONPATH="Your-Location"/xespresso:$PYTHONPATH
export ASE_ESPRESSO_COMMAND="/path/to/PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo"
export ESPRESSO_PSEUDO="/path/to/pseudo"

Run PACKAGE.x jobs.
"""


from ase import io
from ase.calculators.calculator import FileIOCalculator, CalculationFailed, equal, compare_atoms, PropertyNotPresent
from xespresso.xio import write_espresso_in, read_espresso_input, read_espresso_asei, write_espresso_asei, get_atomic_species, get_atomic_constraints
from ase.io.espresso import read_espresso_in
import os
import shutil
import numpy as np
from datetime import datetime
import copy
import logging
import sys
import subprocess
import warnings
from pprint import pprint

logging.basicConfig(stream=sys.stdout,
                    format=('%(levelname)-8s '
                            '[%(funcName)-20s]: %(message)s'),
                    level=logging.INFO)

logger = logging.getLogger(__name__)


error_template = 'Property "%s" not available. Please try running Quantum\n' \
                 'Espresso first by calling Atoms.get_potential_energy().'

warn_template = 'Property "%s" is None. Typically, this is because the ' \
                'required information has not been printed by Quantum ' \
                'Espresso at a "low" verbosity level (the default). ' \
                'Please try running Quantum Espresso with "high" verbosity.'

pseudo_gropus = {
    'SSSP_1.1.2_PBE_efficiency': {'CL': 'cl_pbe_v1.4.uspp.F.UPF', 'TA': 'Ta_pbe_v1.uspp.F.UPF', 'HE': 'He_ONCV_PBE-1.0.oncvpsp.upf', 'OS': 'Os_pbe_v1.2.uspp.F.UPF', 'GE': 'ge_pbe_v1.4.uspp.F.UPF', 'S': 's_pbe_v1.4.uspp.F.UPF', 'CR': 'cr_pbe_v1.5.uspp.F.UPF', 'ND': 'Nd.GGA-PBE-paw-v1.0.UPF', 'LU': 'Lu.GGA-PBE-paw-v1.0.UPF', 'ZN': 'Zn_pbe_v1.uspp.F.UPF', 'MO': 'Mo_ONCV_PBE-1.0.oncvpsp.upf', 'FE': 'Fe.pbe-spn-kjpaw_psl.0.2.1.UPF', 'AR': 'Ar_ONCV_PBE-1.1.oncvpsp.upf', 'CO': 'Co_pbe_v1.2.uspp.F.UPF', 'AG': 'Ag_ONCV_PBE-1.0.oncvpsp.upf', 'ZR': 'Zr_pbe_v1.uspp.F.UPF', 'SB': 'sb_pbe_v1.4.uspp.F.UPF', 'SC': 'Sc_ONCV_PBE-1.0.oncvpsp.upf', 'TM': 'Tm.GGA-PBE-paw-v1.0.UPF', 'BR': 'br_pbe_v1.4.uspp.F.UPF', 'TI': 'ti_pbe_v1.4.uspp.F.UPF', 'TL': 'Tl_pbe_v1.2.uspp.F.UPF', 'LA': 'La.GGA-PBE-paw-v1.0.UPF', 'H': 'H.pbe-rrkjus_psl.1.0.0.UPF', 'AL': 'Al.pbe-n-kjpaw_psl.1.0.0.UPF', 'B': 'b_pbe_v1.4.uspp.F.UPF', 'C': 'C.pbe-n-kjpaw_psl.1.0.0.UPF', 'TC': 'Tc_ONCV_PBE-1.0.oncvpsp.upf', 'IN': 'In.pbe-dn-rrkjus_psl.0.2.2.UPF', 'LI': 'li_pbe_v1.4.uspp.F.UPF', 'PM': 'Pm.GGA-PBE-paw-v1.0.UPF', 'RB': 'Rb_ONCV_PBE-1.0.oncvpsp.upf', 'PD': 'Pd_ONCV_PBE-1.0.oncvpsp.upf', 'BA': 'Ba.pbe-spn-kjpaw_psl.1.0.0.UPF', 'PO': 'Po.pbe-dn-rrkjus_psl.1.0.0.UPF', 'MG': 'Mg.pbe-n-kjpaw_psl.0.3.0.UPF', 'HF': 'Hf-sp.oncvpsp.upf', 'EU': 'Eu.GGA-PBE-paw-v1.0.UPF', 'I': 'I.pbe-n-kjpaw_psl.0.2.UPF', 'K': 'K.pbe-spn-kjpaw_psl.1.0.0.UPF', 'ER': 'Er.GGA-PBE-paw-v1.0.UPF', 'F': 'f_pbe_v1.4.uspp.F.UPF', 'RN': 'Rn.pbe-dn-kjpaw_psl.1.0.0.UPF', 'CE': 'Ce.GGA-PBE-paw-v1.0.UPF', 'RE': 'Re_pbe_v1.2.uspp.F.UPF', 'O': 'O.pbe-n-kjpaw_psl.0.1.UPF', 'KR': 'Kr_ONCV_PBE-1.0.oncvpsp.upf', 'CS': 'Cs_pbe_v1.uspp.F.UPF', 'PR': 'Pr.GGA-PBE-paw-v1.0.UPF', 'GD': 'Gd.GGA-PBE-paw-v1.0.UPF', 'CU': 'Cu_pbe_v1.2.uspp.F.UPF', 'XE': 'Xe_ONCV_PBE-1.1.oncvpsp.upf', 'P': 'P.pbe-n-rrkjus_psl.1.0.0.UPF', 'YB': 'Yb.GGA-PBE-paw-v1.0.UPF', 'CA': 'Ca_pbe_v1.uspp.F.UPF', 'AS': 'As.pbe-n-rrkjus_psl.0.2.UPF', 'NI': 'ni_pbe_v1.4.uspp.F.UPF', 'MN': 'mn_pbe_v1.5.uspp.F.UPF', 'BE': 'be_pbe_v1.4.uspp.F.UPF', 'PT': 'pt_pbe_v1.4.uspp.F.UPF', 'RU': 'Ru_ONCV_PBE-1.0.oncvpsp.upf', 'Y': 'Y_pbe_v1.uspp.F.UPF', 'DY': 'Dy.GGA-PBE-paw-v1.0.UPF', 'W': 'W_pbe_v1.2.uspp.F.UPF', 'TE': 'Te_pbe_v1.uspp.F.UPF', 'V': 'v_pbe_v1.4.uspp.F.UPF', 'IR': 'Ir_pbe_v1.2.uspp.F.UPF', 'SR': 'Sr_pbe_v1.uspp.F.UPF', 'SI': 'Si.pbe-n-rrkjus_psl.1.0.0.UPF', 'N': 'N.pbe-n-radius_5.UPF', 'NE': 'Ne_ONCV_PBE-1.0.oncvpsp.upf', 'NB': 'Nb.pbe-spn-kjpaw_psl.0.3.0.UPF', 'SE': 'Se_pbe_v1.uspp.F.UPF', 'AU': 'Au_ONCV_PBE-1.0.oncvpsp.upf', 'SM': 'Sm.GGA-PBE-paw-v1.0.UPF', 'TB': 'Tb.GGA-PBE-paw-v1.0.UPF', 'NA': 'na_pbe_v1.5.uspp.F.UPF', 'BI': 'Bi_pbe_v1.uspp.F.UPF', 'RH': 'Rh_ONCV_PBE-1.0.oncvpsp.upf', 'GA': 'Ga.pbe-dn-kjpaw_psl.1.0.0.UPF', 'PB': 'Pb.pbe-dn-kjpaw_psl.0.2.2.UPF', 'HO': 'Ho.GGA-PBE-paw-v1.0.UPF', 'CD': 'Cd.pbe-dn-rrkjus_psl.0.3.1.UPF', 'SN': 'Sn_pbe_v1.uspp.F.UPF', 'HG': 'Hg_ONCV_PBE-1.0.oncvpsp.upf'},
    'SSSP_1.1.2_PBE_precision': {'TA': 'Ta_pbe_v1.uspp.F.UPF', 'HE': 'He_ONCV_PBE-1.0.oncvpsp.upf', 'MG': 'mg_pbe_v1.4.uspp.F.UPF', 'OS': 'Os_pbe_v1.2.uspp.F.UPF', 'GE': 'ge_pbe_v1.4.uspp.F.UPF', 'S': 's_pbe_v1.4.uspp.F.UPF', 'CR': 'cr_pbe_v1.5.uspp.F.UPF', 'ND': 'Nd.GGA-PBE-paw-v1.0.UPF', 'LU': 'Lu.GGA-PBE-paw-v1.0.UPF', 'ZN': 'Zn_pbe_v1.uspp.F.UPF', 'MO': 'Mo_ONCV_PBE-1.0.oncvpsp.upf', 'FE': 'Fe.pbe-spn-kjpaw_psl.0.2.1.UPF', 'AR': 'Ar_ONCV_PBE-1.1.oncvpsp.upf', 'CO': 'Co_pbe_v1.2.uspp.F.UPF', 'AG': 'Ag_ONCV_PBE-1.0.oncvpsp.upf', 'ZR': 'Zr_pbe_v1.uspp.F.UPF', 'SB': 'sb_pbe_v1.4.uspp.F.UPF', 'TM': 'Tm.GGA-PBE-paw-v1.0.UPF', 'BR': 'br_pbe_v1.4.uspp.F.UPF', 'TI': 'ti_pbe_v1.4.uspp.F.UPF', 'TL': 'Tl_pbe_v1.2.uspp.F.UPF', 'LA': 'La.GGA-PBE-paw-v1.0.UPF', 'AL': 'Al.pbe-n-kjpaw_psl.1.0.0.UPF', 'C': 'C.pbe-n-kjpaw_psl.1.0.0.UPF', 'TC': 'Tc_ONCV_PBE-1.0.oncvpsp.upf', 'IN': 'In.pbe-dn-rrkjus_psl.0.2.2.UPF', 'LI': 'li_pbe_v1.4.uspp.F.UPF', 'PM': 'Pm.GGA-PBE-paw-v1.0.UPF', 'RB': 'Rb_ONCV_PBE-1.0.oncvpsp.upf', 'PD': 'Pd_ONCV_PBE-1.0.oncvpsp.upf', 'BA': 'Ba.pbe-spn-kjpaw_psl.1.0.0.UPF', 'PO': 'Po.pbe-dn-rrkjus_psl.1.0.0.UPF', 'HF': 'Hf-sp.oncvpsp.upf', 'EU': 'Eu.GGA-PBE-paw-v1.0.UPF', 'H': 'H_ONCV_PBE-1.0.oncvpsp.upf', 'I': 'I.pbe-n-kjpaw_psl.0.2.UPF', 'K': 'K.pbe-spn-kjpaw_psl.1.0.0.UPF', 'ER': 'Er.GGA-PBE-paw-v1.0.UPF', 'RN': 'Rn.pbe-dn-kjpaw_psl.1.0.0.UPF', 'B': 'B_pbe_v1.01.uspp.F.UPF', 'CE': 'Ce.GGA-PBE-paw-v1.0.UPF', 'RE': 'Re_pbe_v1.2.uspp.F.UPF', 'O': 'O.pbe-n-kjpaw_psl.0.1.UPF', 'KR': 'Kr_ONCV_PBE-1.0.oncvpsp.upf', 'CS': 'Cs_pbe_v1.uspp.F.UPF', 'PR': 'Pr.GGA-PBE-paw-v1.0.UPF', 'GD': 'Gd.GGA-PBE-paw-v1.0.UPF', 'XE': 'Xe_ONCV_PBE-1.1.oncvpsp.upf', 'P': 'P.pbe-n-rrkjus_psl.1.0.0.UPF', 'N': 'N.oncvpsp.upf', 'YB': 'Yb.GGA-PBE-paw-v1.0.UPF', 'BE': 'Be_ONCV_PBE-1.0.oncvpsp.upf', 'CA': 'Ca_pbe_v1.uspp.F.UPF', 'AS': 'As.pbe-n-rrkjus_psl.0.2.UPF', 'NI': 'ni_pbe_v1.4.uspp.F.UPF', 'MN': 'mn_pbe_v1.5.uspp.F.UPF', 'SC': 'Sc.pbe-spn-kjpaw_psl.0.2.3.UPF', 'RU': 'Ru_ONCV_PBE-1.0.oncvpsp.upf', 'Y': 'Y_pbe_v1.uspp.F.UPF', 'DY': 'Dy.GGA-PBE-paw-v1.0.UPF', 'F': 'F.oncvpsp.upf', 'W': 'W_pbe_v1.2.uspp.F.UPF', 'TE': 'Te_pbe_v1.uspp.F.UPF', 'V': 'v_pbe_v1.4.uspp.F.UPF', 'PT': 'Pt.pbe-spfn-rrkjus_psl.1.0.0.UPF', 'IR': 'Ir_pbe_v1.2.uspp.F.UPF', 'SR': 'Sr_pbe_v1.uspp.F.UPF', 'SI': 'Si.pbe-n-rrkjus_psl.1.0.0.UPF', 'CL': 'Cl.pbe-n-rrkjus_psl.1.0.0.UPF', 'NE': 'Ne_ONCV_PBE-1.0.oncvpsp.upf', 'NB': 'Nb.pbe-spn-kjpaw_psl.0.3.0.UPF', 'SE': 'Se_pbe_v1.uspp.F.UPF', 'AU': 'Au_ONCV_PBE-1.0.oncvpsp.upf', 'CU': 'Cu_ONCV_PBE-1.0.oncvpsp.upf', 'SM': 'Sm.GGA-PBE-paw-v1.0.UPF', 'TB': 'Tb.GGA-PBE-paw-v1.0.UPF', 'BI': 'Bi_pbe_v1.uspp.F.UPF', 'RH': 'Rh_ONCV_PBE-1.0.oncvpsp.upf', 'GA': 'Ga.pbe-dn-kjpaw_psl.1.0.0.UPF', 'PB': 'Pb.pbe-dn-kjpaw_psl.0.2.2.UPF', 'HO': 'Ho.GGA-PBE-paw-v1.0.UPF', 'CD': 'Cd.pbe-dn-rrkjus_psl.0.3.1.UPF', 'SN': 'Sn_pbe_v1.uspp.F.UPF', 'NA': 'Na_ONCV_PBE-1.0.oncvpsp.upf', 'HG': 'Hg_ONCV_PBE-1.0.oncvpsp.upf'},
}

class Espresso(FileIOCalculator):
    implemented_properties = ['energy', 'forces', 'stress', 'magmoms', 'time']
    command = 'PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo'
    discard_results_on_any_change = False

    def __init__(self, label='xespresso', prefix=None, atoms=None, package='pw', parallel='',
                 queue=None, debug=False, 
                 pseudo_group = None,
                 **kwargs):
        """
        Accepts all the options for pw.x as given in the QE docs, plus some
        additional options:

        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE
        queue: dict
            A dictionary with parameters for job submission, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
        package: str
            Choose the quantum espresso package: pw, dos, projwfc, band, pp, ph, ..
            For NEB calculation, please use neb.NEBEspresso module.
            Calculaiton use phonon is not implemented yet.
        parallel: str
            A str which control the parallelization parameters: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).

        General working procedure is as follows:
            1. Perform a regular self-consistent calculation:
              >>> calc = Espresso(input_data=input_data, ...)
              >>> atoms.set_calculator(calc)
              >>> atoms.get_potential_energy()
            2. post calculation can be made as follows:
              >>> calc.post('package' = 'dos', **kwargs)
              >>> calc.post('package' = 'pp', **kwargs)
            3. non-self-consistent calculation can be made as follows:
              >>> calc.nscf(calculation = 'nscf', kpts=(4, 4, 4))
              >>> calc.nscf_calculate()
        """
        print("{0:=^60}".format(package))
        if debug:
            logger.setLevel(debug)
        self.set_label(label, prefix)
        self.scf_directory = None
        self.scf_parameters = None
        self.scf_results = None
        if atoms:
            self.atoms = atoms
        if pseudo_group is not None:
            kwargs['pseudopotentials'] = self.find_pseudopotentials(pseudo_group)
            kwargs['pseudo_dir'] = os.path.join(os.environ['ESPRESSO_PSEUDO'], pseudo_group)
        kwargs = self.check_input(kwargs, prefix=self.prefix)
        self.ase_parameters = kwargs
        FileIOCalculator.__init__(self, restart=self.directory,
                                  label=self.label, atoms=atoms, **kwargs)
        
        self.queue = queue
        self.parallel = parallel
        self.package = package
        self.parallel = parallel
        self.debug = debug
        
        # self.discard_results_on_any_change = False

    def find_pseudopotentials(self, pseudo_group):
        """_summary_

        Args:
            pseudo_group (_type_): _description_
        """
        elements = set(self.atoms.get_chemical_symbols())
        pseudopotentials = {}
        for ele in elements:
            pseudopotentials[ele] = pseudo_gropus[pseudo_group][ele.upper()]
        return pseudopotentials

    def set_label(self, label, prefix):
        '''
        set directory and prefix from label
        '''
        self.directory = label
        if not prefix:
            self.prefix = os.path.split(label)[1]
        else:
            self.prefix = prefix
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        self.label = os.path.join(self.directory, self.prefix)
        self.pwi = os.path.join(self.directory, '%s.pwi' % self.prefix)
        self.pwo = os.path.join(self.directory, '%s.pwo' % self.prefix)
        self.asei = os.path.join(self.directory, '%s.asei' % self.prefix)
        self.asei_temp = os.path.join(
            self.directory, '.%s.asei_temp' % self.prefix)
        self.save_directory = os.path.join(
            self.directory, '%s.save' % self.prefix)
        logger.debug('Directory: %s' % (self.directory))
        logger.debug('Prefix: %s' % (self.prefix))

    def check_pseudopotentials(self, pseudopotentials):
        """
        """
        # pseudopotentials
        if 'species' not in self.atoms.arrays:
            all_species = set(self.atoms.get_chemical_symbols())
        else:
            all_species = set(self.atoms.arrays['species'])
        new_pseudopotentials = {}
        for species in all_species:
            assert species in pseudopotentials, '\n  Species: %s, does not have a pseudopotentials!' % (
                species)
            new_pseudopotentials[species] = pseudopotentials[species]

    @classmethod
    def check_input(cls, kwargs, prefix='pw', package='PW'):
        """
        Chekc input paramter type.
        set parameters for pw.x
        all other parameters are stored in 'input_data'
        """
        from xespresso.xio import sort_qe_input, check_qe_input
        sorted_parameters, unuse_parameters = sort_qe_input(kwargs)
        check_qe_input(sorted_parameters['input_data'])
        if unuse_parameters:
            logger.debug(
                'Warnning: parameter %s not used, please check.' % unuse_parameters)
        sorted_parameters['input_data']['CONTROL']['prefix'] = prefix
        sorted_parameters['input_data']['CONTROL']['verbosity'] = 'high'
        # pprint(sorted_parameters)
        return sorted_parameters

    def read(self, label):
        """Read the files in a calculation if they exist.
        This function reads self.parameters and atoms.
        """
        # Else read a regular calculation. we start with reading stuff
        # that is independent of the calculation state.
        self.directory = label
        self.restart_atoms = None
        self.restart_parameters = {}
        if os.path.exists(label) and os.path.exists(self.asei):
            logger.debug('Try to read information from folder: %s.' % (label))
            try:
                # atoms, parameters = self.read_xml_file()
                atoms, parameters = read_espresso_asei(self.asei)
                self.restart_atoms = atoms
                self.restart_parameters = copy.deepcopy(parameters)
                # self.parameters = copy.deepcopy(parameters)
                self.read_results()
            except Exception as e:
                logger.debug(
                    'Directory: %s exist, faild to read. %s' % (label, e))
        else:
            logger.debug(
                'Directory %s is not a espresso folder, start a new calculation:' % (label))

    def set(self, **kwargs):
        """
        """
        from xespresso.input_parameters import restart_ignore
        self.changed_parameters, self.igonre_parameters = self.compare_parameters(
            self.restart_parameters, self.ase_parameters, ignore=restart_ignore['PW'])
        self.parameters = kwargs
        if self.discard_results_on_any_change and self.changed_parameters:
            self.reset()
        return self.changed_parameters

    def check_state(self, atoms, tol=1e-4):
        """Check for any system changes since last calculation."""
        if not os.path.exists(self.asei):
            return True
        if isinstance(self.restart_atoms, list):
            if len(self.restart_atoms) != len(atoms):
                return True
            for i in range(len(atoms)):
                if compare_atoms(self.restart_atoms[i], atoms[i], tol=tol,
                                 excluded_properties=set(self.ignored_changes)):
                    logger.debug('Atoms changed, run a new calculation')
                    return True
        else:
            if compare_atoms(self.restart_atoms, atoms, tol=tol,
                             excluded_properties=set(self.ignored_changes)):
                logger.debug('Atoms changed, run a new calculation')
                return True
        # if self.igonre_parameters:
            # logger.debug('Parameter: %s are ignored, results are not affected.'%self.igonre_parameters)
        if self.changed_parameters:
            logger.debug('Parameter: %s changed, results are affected' %
                         self.changed_parameters)
            return True
        logger.debug('Same geometry and parameters, use previous results.')
        converged, meg = self.read_convergence()
        if converged > 0:
            logger.debug('Not converged: %s' % meg)
            return True
        return False

    def compare_parameters(self, para1, para2, ignore=[]):
        """
        """
        from xespresso.input_parameters import qe_namespace, default_parameters, restart_ignore

        changed_parameters = []
        igonre_parameters = []
        if not para1:
            changed_parameters = ['all']
            return changed_parameters, igonre_parameters
        default_parameters = default_parameters['PW']
        # pseudopotentials
        key = 'pseudopotentials'
        try:
            for species, value in para1[key].items():
                if value != para2[key][species]:
                    changed_parameters.append(key)
                    continue
        except:
            changed_parameters.append(key)
        # kpts
        key = 'kpts'
        try:
            if para1[key] != para2[key]:
                changed_parameters.append(key)
        except:
            changed_parameters.append(key)
        # input_data
        for section, paras in para1['input_data'].items():
            if section == 'INPUT_NTYP':
                changed_parameters1, igonre_parameters1 = self.compare_dict(
                    para1['input_data'][section], para2['input_data'][section], restart_ignore['PW'])
            else:
                changed_parameters1, igonre_parameters1 = self.compare_dict(
                    para1['input_data'][section], para2['input_data'][section], restart_ignore['PW'], default=default_parameters[section])
            changed_parameters.extend(changed_parameters1)
            igonre_parameters.extend(igonre_parameters1)
        return changed_parameters, igonre_parameters

    def compare_dict(self, dict1, dict2, ignore=[], default=None):
        igonre_parameters = []
        changed_parameters = []
        keys = set(list(dict1.keys()) + list(dict2.keys()))
        print
        for key in keys:
            if key in ignore:
                igonre_parameters.append(key)
            elif key not in dict1:
                changed_parameters.append(key)
            elif key not in dict2:
                if default and self.compare_value(dict1[key], default[key]):
                    continue
                changed_parameters.append(key)
            elif not self.compare_value(dict1[key], dict2[key]):
                changed_parameters.append(key)
        return changed_parameters, igonre_parameters

    def compare_value(self, v1, v2, tol=1e-5):
        """
        """
        if isinstance(v1, str):
            if v1.upper() != v2.upper():
                return False
        elif isinstance(v1, dict):
            changed_parameters, igonre_parameters = self.compare_dict(v1, v2)
            if changed_parameters:
                return False
            for key, value in v1.items():
                if not self.compare_value(value, v2[key]):
                    return False
        elif isinstance(v1, bool):
            if v1 != v2:
                return False
        else:
            if abs(v1 - v2) > tol:
                return False
        return True

    def read_results(self):
        '''
        get atomic species
        '''
        pwo = os.path.join(self.directory, '%s.pwo' % self.prefix)
        pwi = os.path.join(self.directory, '%s.pwi' % self.prefix)
        convergence, meg = self.read_convergence()
        if convergence != 0:
            logger.debug('Not converged. %s' % (meg))
        try:
            atoms = read_espresso_in(pwi)
            self.results['atoms'] = atoms
        except Exception as e:
            logger.debug('Read input: %s, failed! %s' % (pwi, e))
        try:
            output = io.read(pwo)
            atomic_species = get_atomic_species(pwo)
            constraints = get_atomic_constraints(pwo, len(output))
            output.set_constraint(None)
            output.set_constraint(constraints)
            if atomic_species:
                output.new_array('species', np.array(
                    atomic_species, dtype='U20'))
            self.calc = output.calc
            self.results = output.calc.results
            self.results['atoms'] = output
            self.efermi = self.get_fermi_level()
            # self.nspins = self.get_number_of_spins()
            logger.debug('Read result successfully!')
        except Exception as e:
            logger.debug('Read output: %s, failed! %s' % (pwo, e))
        self.results['convergence'] = convergence
        # logger.debug('Read result failed!')
        # pwos = [file for file in os.listdir(self.directory) if pwo in file]
        # output = None
        # for pwo in pwos:
        # atomic_species = None
        # pwo = os.path.join(self.directory, pwo)
        # atomic_species = get_atomic_species(pwo)

    def write_input(self, atoms, properties=None, system_changes=None):
        from xespresso.scheduler import set_queue
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        set_queue(self)
        write_espresso_in(self.label + '.pwi', atoms, **self.parameters)
        write_espresso_asei(self.label + '.asei', atoms, self.parameters)
        logger.debug('Write input successfully')

    def check_xml_file(self):
        '''
        If data-file-schema.xml is readable, then xml_end = True
        '''
        self.save_directory = os.path.join(
            self.directory, '%s.save' % self.prefix)
        if not os.path.exists(self.save_directory):
            return False
        xml0 = os.path.join(self.save_directory, 'data-file-schema.xml')
        xmls = [file for file in os.listdir(
            self.save_directory) if 'data-file-schema.xml' in file]
        goodxml = []
        for xml in xmls:
            xml = os.path.join(self.save_directory, xml)
            with open(xml, 'r') as f:
                lines = f.readlines()
                if len(lines) < 1:
                    continue
                if '</qes:espresso>' in lines[-1]:
                    goodxml.append(xml)
        xml_end = False
        nxml = len(goodxml)
        # print(nxml, goodxml, xml0)
        if nxml > 0:
            if xml0 not in goodxml:
                shutil.copy(goodxml[-1], xml0)
            xml_end = True
        return xml_end

    def read_xml_file(self):
        from xespresso.utils.xml_parser import xml_parser
        if not self.check_xml_file():
            logger.debug('xml file not finished.')
        #
        xmlfile = os.path.join(self.save_directory, 'data-file-schema.xml')
        if not os.path.exists(xmlfile):
            return None, {}
        try:
            atoms, xml_input = xml_parser(xmlfile)
            logger.debug(
                'Read geometry and parameters from xml file successfully.')
        except:
            logger.debug('Read xml file failed.')
            return None, {}
        return atoms, xml_input

    def read_convergence(self):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        '-1': ' manual cancelled',
        '1': 'Maximum CPU time exceeded', convergence NOT achieved after n iterations
             Then, change the parameter: 1) mixing_beta or 2) degauss
        '2': 'manual restart', 'Pending or not submit',
        '3': 'other errors',
        '4': 'unknow error',
        }
        '''
        # read the error file from queue
        # if self.queue:
        #     errfile = self.label + '.err'
        #     if not os.path.exists(errfile):
        #         logger.debug('%s not exists'%errfile)
        #         # return 2, 'Pending or not submit'
        #     else:
        #         errs = ['RESTART', 'pw.x', 'out-of-memory',
        #                 'NODE FAILURE', 'TIME LIMIT', 'COMMUNICATION',
        #                 'segmentation fault', 'PARSE_ERR', 'mpirun']
        #         cancelled = False
        #         with open(errfile, 'r') as f:
        #             lines = f.readlines()
        #             for line in lines:
        #                 if 'CANCELLED' in line:
        #                     cancelled = True
        #                 for err in errs:
        #                     if err in line:
        #                         logger.debug('Need restart')
        #                         return 3, line
        #             # cancelled by owner
        #             if len(lines) > 0:
        #                 if cancelled:
        #                     logger.debug('Cancelled')
        #                     return -1, lines[0]
        # no error from queue, or job not run in the queue
        output = self.label + '.pwo'
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return 3, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0:
                return 1, 'pwo file has nothing'
            stime = lines[1].split('starts on')[1]
            nlines = len(lines)
            n = min([200, nlines])
            lastlines = lines[-n:-1]
            for line in lastlines:
                if line.rfind('too many bands are not converged') > -1:
                    logger.debug('Need restart')
                    return 1, 'Reason: %s' % (line)
                if line.rfind('convergence NOT achieved after') > -1:
                    logger.debug('Need restart: %s' % line)
                    return 1, 'Reason: %s' % (line)
                if line.rfind('Maximum CPU time exceeded') > -1:
                    logger.debug('Need restart')
                    return 2, 'Reason: %s' % (line)
                if line.rfind('JOB DONE.') > -1:
                    logger.debug('JOB DONE')
                    return 0, line
            logger.debug('Not converged, %s' % lastlines[-1])
        return 4, line

    def run(self, atoms=None, restart=False):
        '''
        run and restart
        '''
        if atoms is not None:
            self.atoms = atoms
        icount = 1
        converged = 4
        meg0 = 'None'
        self.atoms.calc = self
        if not restart:
            try:
                logger.debug('Run step: %s' % icount)
                self.atoms.get_potential_energy()
                converged, meg0 = self.read_convergence()
            except Exception as e:
                logger.debug('Not converge: %s' % e)
                converged, meg0 = self.read_convergence()
                icount += 1
                if converged == 0:
                    converged = 4
        # print(converged, xml_end, meg0)
        # restart 0, 1, 2
        exitfile = os.path.join(self.directory, 'EXIT')
        while converged > 0 or restart == 2:
            if os.path.exists(exitfile):
                logger.debug('Exit file exists!')
                os.remove(exitfile)
                sys.exit()
            restart = 1
            self.backup_file('%s.pwo' % self.prefix, directory=self.directory)
            self.backup_file('data-file-schema.xml',
                             directory=self.save_directory)
            logger.debug('Run step: %s' % icount)
            xml_end = self.check_xml_file()
            if xml_end:
                self.parameters['input_data']['CONTROL']['restart_mode'] = 'restart'
            else:
                self.parameters['input_data']['CONTROL']['restart_mode'] = 'from_scratch'
            if converged == 1:
                if 'electron_maxstep' in self.parameters['input_data']['ELECTRONS']:
                    electron_maxstep = int(
                        self.parameters['input_data']['ELECTRONS']['electron_maxstep'])
                else:
                    electron_maxstep = 100
                if 'mixing_beta' in self.parameters['input_data']['ELECTRONS']:
                    mixing_beta = float(
                        self.parameters['input_data']['ELECTRONS']['mixing_beta'])
                else:
                    mixing_beta = 0.7
                if electron_maxstep < 100:
                    self.parameters['input_data']['ELECTRONS']['electron_maxstep'] = electron_maxstep + 50
                    logger.debug('electron_maxstep change from %s to %s' % (
                        electron_maxstep, electron_maxstep + 50))
                self.parameters['input_data']['ELECTRONS']['mixing_beta'] = mixing_beta/2.0
                logger.debug('mixing_beta change from %s to %s' %
                             (mixing_beta, mixing_beta/2.0))
            try:
                icount += 1
                atoms.get_potential_energy()
                converged, meg = self.read_convergence()
            except Exception as e:
                logger.debug('Not converge: %s' % e)
                converged, meg = self.read_convergence()
                meg = str(e)
                restart = 2
                # if converged == 0:
                converged = 4
            logger.debug('%s, %s' % (converged, meg))
            if meg == meg0 and converged != 2:
                logger.debug(
                    'Exit! Maybe deadblock! Same error message! %s' % meg)
                sys.exit()
            meg0 = meg

    def backup_file(self, src, directory='.'):
        '''
        compare files
        backup
        '''
        import filecmp
        tnow = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        new_src = os.path.join(directory, '%s-%s' % (tnow, src))
        src = os.path.join(directory, src)
        if not os.path.exists(src):
            return 0
        ftype = src.split('.')[-1]
        flag = True
        for dst in os.listdir(directory):
            dst = os.path.join(directory, dst)
            if dst == src:
                continue
            if dst.endswith(ftype):
                if filecmp.cmp(dst, src):
                    flag = False
        # print('backup files: ', flag, src, new_src)
        if flag:
            shutil.copy(src, new_src)
        return 0

    def plot_phdos(self, fldos=None, ax=None, output=None):
        '''
        '''
        import matplotlib.pyplot as plt
        if fldos is None:
            fldos = self.prefix
        phdos = np.loadtxt(self.directory+'/%s.phdos' % fldos)
        self.phdos = phdos
        if ax is None:
            fig, ax = plt.subplots(figsize=(6, 3))
            # ax = plt.gca()
        ax.plot(self.phdos[:, 0], self.phdos[:, 1], linewidth=0.7)
        ax.set_xlabel('Frequency (cm^-1)')
        ax.set_ylabel('DOS (a.u.)')
        if output is not None:
            plt.savefig('%s' % output)
        return ax

    def get_work_function(self, ax=None, inpfile='potential.cube', output=None, shift=False):
        import matplotlib.pyplot as plt
        from ase.io.cube import read_cube_data
        from ase.units import create_units
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()
        units = create_units('2006')
        #
        filename = os.path.join(self.directory, inpfile)
        data, atoms = read_cube_data(filename)
        data = data * units['Ry']
        ef = self.get_fermi_level()
        # x, y, z, lp = calc.get_local_potential()
        nx, ny, nz = data.shape
        axy = np.array([np.average(data[:, :, z]) for z in range(nz)])
        # setup the x-axis in realspace
        uc = atoms.get_cell()
        xaxis = np.linspace(0, uc[2][2], nz)
        if shift:
            ax.plot(xaxis, axy - ef)
            ef = 0
        else:
            ax.plot(xaxis, axy)
            ax.plot([min(xaxis), max(xaxis)], [ef, ef], 'k:')
        ax.set_xlabel('Position along z-axis ($\AA$)')
        ax.set_ylabel('Potential (eV)')
        if output:
            plt.savefig('%s' % output)
        # plt.show()
        atoms = self.results['atoms']
        pos = max(atoms.positions[:, 2] + 3)
        ind = (xaxis > pos) & (xaxis < pos + 3)
        wf = np.average(axy[ind]) - ef
        print('min: %s, max: %s' % (pos, pos + 3))
        print('The workfunction is {0:1.2f} eV'.format(wf))

    def get_bader_charge(self, inpfile=None):
        '''
        '''
        from ase.io.bader import attach_charges
        if not inpfile:
            inpfile = '%s.cube' % self.prefix
        command = 'bader %s' % inpfile
        print(command)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err
        acf = os.path.join(self.directory, 'ACF.dat')
        attach_charges(self.results['atoms'], acf)

    def read_time(self, ):
        '''
        '''
        pwo = os.path.join(self.directory, self.prefix + '.pwo')
        with open(pwo, 'r') as f:
            ts = 0
            lines = f.readlines()
            for line in lines[::-1]:
                if 'PWSCF' in line and 'WALL' in line:
                    t = line.split('CPU')[-1].split('WALL')[0]
                    if 's' in t:
                        ts = float(t.split('m')[-1].split('s')[0])
                    if 'm' in t:
                        tm = float(t.split('h')[-1].split('m')[0])
                        ts += tm*60
                    if 'h' in t:
                        th = float(t.split('h')[0])
                        ts += th*3600
                    break
                self.results['time'] = ts
        return ts

    def get_time(self, ):
        t = 0
        filename = os.path.join(self.directory, '%s.pwo' % self.prefix)
        with open(filename) as f:
            lines = f.readlines()
            for line in lines[-200:]:
                if 'init_run     :' in line or 'electrons    :' in line:
                    t += float(line.split('CPU')[1].split('s WALL')[0])
        self.results['time'] = t
        return t

    def clean(self):
        '''
        remove wfc, hub files
        '''
        keys = ['.wfc', '.hub']
        files = os.listdir(self.directory)
        for file in files:
            for key in keys:
                if key in file:
                    os.remove(os.path.join(self.directory, file))
        files = os.listdir(self.save_directory)
        # keys = ['wfc']
        # for file in files:
        #     for key in keys:
        #         if key in file:
        #             os.remove(os.path.join(self.save_directory, file))

    def get_fermi_level(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Fermi level')
        return self.calc.get_fermi_level()

    def get_ibz_k_points(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'IBZ k-points')
        ibzkpts = self.calc.get_ibz_k_points()
        if ibzkpts is None:
            warnings.warn(warn_template % 'IBZ k-points')
        return ibzkpts

    def get_k_point_weights(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'K-point weights')
        k_point_weights = self.calc.get_k_point_weights()
        if k_point_weights is None:
            warnings.warn(warn_template % 'K-point weights')
        return k_point_weights

    def get_eigenvalues(self, **kwargs):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Eigenvalues')
        eigenvalues = self.calc.get_eigenvalues(**kwargs)
        if eigenvalues is None:
            warnings.warn(warn_template % 'Eigenvalues')
        return eigenvalues

    def get_number_of_spins(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Number of spins')
        nspins = self.calc.get_number_of_spins()
        if nspins is None:
            warnings.warn(warn_template % 'Number of spins')
        return nspins




from xespresso.post.base import PostCalculation


class EspressoDynmat(PostCalculation):

    package = 'dynmat'

    package_parameters = {'INPUT': ['fildyn', 'q', 'amass', 'asr', 'axis', 'lperm', 'lplasma', 'filout',
                                    'fileig', 'filmol', 'filxsf', 'loto_2d', 'el_ph_nsig', 'el_ph_sigma']
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)




from xespresso.post.base import PostCalculation


class EspressoBands(PostCalculation):

    package = 'bands'
    package_parameters = {'BANDS': ['prefix', 'outdir', 'filband', 'spin_component', 'lsigma',
                                    'lp', 'filp', 'lsym', 'no_overlap', 'plot_2d', 'firstk', 'lastk']
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)




from xespresso.post.base import PostCalculation


class EspressoHp(PostCalculation):

    package = 'hp'
    package_parameters = {'INPUTHP': ['prefix', 'outdir', 'iverbosity', 'max_seconds', 'nq1', 'nq2',
                                      'nq3', 'skip_equivalence_q', 'determine_num_pert_only',
                                      'find_atpert', 'docc_thr', 'skip_type', 'equiv_type',
                                      'perturb_only_atom', 'start_q', 'last_q', 'sum_pertq',
                                      'compute_hp', 'conv_thr_chi', 'thresh_init', 'ethr_nscf',
                                      'niter_max', 'alpha_mix(i)', 'nmix', 'num_neigh', 'lmin', 'rmax', ]
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)





restart_ignore = {'PW': ['max_seconds', 'nstep', 'restart_mode',
                         'electron_maxstep',
                         'scf_must_converge',
                         'mixing_mode',
                         'mixing_beta',
                         'mixing_ndim',
                         'mixing_fixed_ns',
                         'diagonalization',
                         'diago_cg_maxiter',
                         'diago_david_ndim',
                         'diago_full_acc',
                         'startingpot',
                         'startingwfc',
                         'tqr',
                         'real_space',
                         'title',
                         'disk_io',
                         'wf_collect',
                         'outdir',
                         'pseudo_dir',
                         #
                         'diago_thr_init',
                         'tprnfor',
                         ]}

default = {
    'PW': {'disk_io': {'scf': 'low', 'nscf': 'medium', 'bands': 'medium', 'relax': 'medium', 'md': 'medium', 'vc-relax': 'medium', 'vc-md': 'medium'},
           'nstep': {'scf': 1, 'nscf': 1, 'bands': 1, 'relax': 50, 'md': 50, 'vc-relax': 50, 'vc-md': 50},
           'pseudo_dir': '$HOME/espresso/pseudo/',
           },
}

Ha2Ry = {
    'ecutwfc',
    'ecutrho',
    'ecutfock',
    # 'degauss',  # not sure
    'etot_conv_thr',
    'forc_conv_thr',
}

qe_namespace = {'BANDS': {'BANDS': {'filband': ['CHARACTER', []],
                                    'filp': ['CHARACTER', []],
                                    'firstk, lastk': ['INTEGER', []],
                                    'lp': ['LOGICAL', []],
                                    'lsigma(i), i=1,3': ['LOGICAL', []],
                                    'lsym': ['LOGICAL', []],
                                    'no_overlap': ['LOGICAL', []],
                                    'outdir': ['CHARACTER', []],
                                    'plot_2d': ['LOGICAL', []],
                                    'prefix': ['CHARACTER', []],
                                    'spin_component': ['INTEGER', []]}},
                'DOS': {'DOS': {'DeltaE': ['REAL', []],
                                'Emin, Emax': ['REAL', []],
                                'bz_sum': ['CHARACTER',
                                           ['smearing',
                                            'tetrahedra',
                                            'tetrahedra_lin',
                                            'tetrahedra_opt']],
                                'degauss': ['REAL', []],
                                'fildos': ['CHARACTER', []],
                                'ngauss': ['INTEGER', []],
                                'outdir': ['CHARACTER', []],
                                'prefix': ['CHARACTER', []]}},
                'NEB': {'PATH': {'CI_scheme': ['CHARACTER', ['no-CI', 'auto', 'manual']],
                                 'ds': ['REAL', []],
                                 'fcp_mu': ['REAL', []],
                                 'fcp_tot_charge_first': ['REAL', []],
                                 'fcp_tot_charge_last': ['REAL', []],
                                 'first_last_opt': ['LOGICAL', []],
                                 'k_max, k_min': ['REAL', []],
                                 'lfcpopt': ['LOGICAL', []],
                                 'minimum_image': ['LOGICAL', []],
                                 'nstep_path': ['INTEGER', []],
                                 'num_of_images': ['INTEGER', []],
                                 'opt_scheme': ['CHARACTER',
                                                ['sd',
                                                 'broyden',
                                                 'broyden2',
                                                 'quick-min',
                                                 'langevin']],
                                 'path_thr': ['REAL', []],
                                 'restart_mode': ['CHARACTER', ['from_scratch', 'restart']],
                                 'string_method': ['CHARACTER', ['neb', 'smd']],
                                 'temp_req': ['REAL', []],
                                 'use_freezing': ['LOGICAL', []],
                                 'use_masses': ['LOGICAL', []]},
                        'ATOMIC_POSITIONS': {},
                        'CLIMBING_IMAGES': {'index1, index2, ... indexN': ['INTEGER', []]}},
                'PP': {'INPUTPP': {'degauss_ldos': ['REAL', []],
                                   'delta_e': ['REAL', []],
                                   'emax': ['REAL', []],
                                   'emin': ['REAL', []],
                                   'filplot': ['CHARACTER', []],
                                   'kband(i), i=1,2': ['INTEGER', []],
                                   'kpoint(i), i=1,2': ['INTEGER', []],
                                   'lsign': ['LOGICAL', []],
                                   'outdir': ['CHARACTER', []],
                                   'plot_num': ['INTEGER', []],
                                   'prefix': ['CHARACTER', []],
                                   'sample_bias': ['REAL', []],
                                   'spin_component': ['INTEGER', []],
                                   'spin_component(i), i=1,2': ['INTEGER', []]},
                       'PLOT': {'e1(i), \n\t\te2(i), \n\t\t \n\t\t  i=1,3': ['REAL', []],
                                'e1(i), \n\t\te2(i), \n\t\te3(i), \n\t\t \n\t\t  i=1,3': ['REAL',
                                                                                          []],
                                'e1(i), i=1,3': ['REAL', []],
                                'fileout': ['CHARACTER', []],
                                'filepp(i), i=1,nfile': ['CHARACTER', []],
                                'iflag': ['INTEGER', []],
                                'interpolation': ['CHARACTER', ['fourier', 'bspline']],
                                'nfile': ['INTEGER', []],
                                'nx': ['INTEGER', []],
                                'nx, ny': ['INTEGER', []],
                                'nx, ny, nz': ['INTEGER', []],
                                'output_format': ['INTEGER', []],
                                'radius': ['REAL', []],
                                'weight(i), i=1,nfile': ['REAL', []],
                                'x0(i), i=1,3': ['REAL', []]}},
                'PROJWFC': {'PROJWFC': {'DeltaE': ['REAL', []],
                                        'Emin, Emax': ['REAL', []],
                                        'degauss': ['REAL', []],
                                        'filpdos': ['CHARACTER', []],
                                        'filproj': ['CHARACTER', []],
                                        'irmax(3,n_proj_boxes)': ['INTEGER', []],
                                        'irmin(3,n_proj_boxes)': ['INTEGER', []],
                                        'kresolveddos': ['LOGICAL', []],
                                        'lbinary_data': ['LOGICAL', []],
                                        'lsym': ['LOGICAL', []],
                                        'lwrite_overlaps': ['LOGICAL', []],
                                        'n_proj_boxes': ['INTEGER', []],
                                        'ngauss': ['INTEGER', []],
                                        'outdir': ['CHARACTER', []],
                                        'pawproj': ['LOGICAL', []],
                                        'plotboxes': ['LOGICAL', []],
                                        'prefix': ['CHARACTER', []],
                                        'tdosinboxes': ['LOGICAL', []]}},
                'PW': {'CELL': {'cell_dofree': ['CHARACTER',
                                                ['all',
                                                 'ibrav',
                                                 'x',
                                                 'y',
                                                 'z',
                                                 'xy',
                                                 'xz',
                                                 'yz',
                                                 'xyz',
                                                 'shape',
                                                 'volume',
                                                 '2Dxy',
                                                 '2Dshape',
                                                 'epitaxial_ab',
                                                 'epitaxial_ac',
                                                 'epitaxial_bc']],
                                'cell_dynamics': ['CHARACTER',
                                                  ['none',
                                                   'sd',
                                                   'damp-pr',
                                                   'damp-w',
                                                   'bfgs',
                                                   'none',
                                                   'pr',
                                                   'w']],
                                'cell_factor': ['REAL', []],
                                'press': ['REAL', []],
                                'press_conv_thr': ['REAL', []],
                                'wmass': ['REAL', []]},
                       'CONTROL': {'calculation': ['CHARACTER',
                                                   ['scf',
                                                    'nscf',
                                                    'bands',
                                                    'relax',
                                                    'md',
                                                    'vc-relax',
                                                    'vc-md']],
                                   'dipfield': ['LOGICAL', []],
                                   'disk_io': ['CHARACTER',
                                               ['high', 'medium', 'low', 'nowf', 'none']],
                                   'dt': ['REAL', []],
                                   'etot_conv_thr': ['REAL', []],
                                   'forc_conv_thr': ['REAL', []],
                                   'gate': ['LOGICAL', []],
                                   'gdir': ['INTEGER', []],
                                   'iprint': ['INTEGER', []],
                                   'lberry': ['LOGICAL', []],
                                   'lelfield': ['LOGICAL', []],
                                   'lfcpopt': ['LOGICAL', []],
                                   'lkpoint_dir': ['LOGICAL', []],
                                   'lorbm': ['LOGICAL', []],
                                   'max_seconds': ['REAL', []],
                                   'nberrycyc': ['INTEGER', []],
                                   'nppstr': ['INTEGER', []],
                                   'nstep': ['INTEGER', []],
                                   'outdir': ['CHARACTER', []],
                                   'prefix': ['CHARACTER', []],
                                   'pseudo_dir': ['CHARACTER', []],
                                   'restart_mode': ['CHARACTER', ['from_scratch', 'restart']],
                                   'tefield': ['LOGICAL', []],
                                   'title': ['CHARACTER', []],
                                   'tprnfor': ['LOGICAL', []],
                                   'tstress': ['LOGICAL', []],
                                   'verbosity': ['CHARACTER', ['high', 'low']],
                                   'wf_collect': ['LOGICAL', []],
                                   'wfcdir': ['CHARACTER', []]},
                       'ELECTRONS': {'adaptive_thr': ['LOGICAL', []],
                                     'conv_thr': ['REAL', []],
                                     'conv_thr_init': ['REAL', []],
                                     'conv_thr_multi': ['REAL', []],
                                     'diago_cg_maxiter': ['INTEGER', []],
                                     'diago_david_ndim': ['INTEGER', []],
                                     'diago_full_acc': ['LOGICAL', []],
                                     'diago_thr_init': ['REAL', []],
                                     'diagonalization': ['CHARACTER',
                                                         ['david',
                                                          'cg',
                                                          'ppcg',
                                                          'paro',
                                                          'ParO']],
                                     'efield': ['REAL', []],
                                     'efield_cart(i), i=1,3': ['REAL', []],
                                     'efield_phase': ['CHARACTER', ['read', 'write', 'none']],
                                     'electron_maxstep': ['INTEGER', []],
                                     'mixing_beta': ['REAL', []],
                                     'mixing_fixed_ns': ['INTEGER', []],
                                     'mixing_mode': ['CHARACTER',
                                                     ['plain', 'TF', 'local-TF']],
                                     'mixing_ndim': ['INTEGER', []],
                                     'real_space': ['LOGICAL', []],
                                     'scf_must_converge': ['LOGICAL', []],
                                     'startingpot': ['CHARACTER', ['atomic', 'file']],
                                     'startingwfc': ['CHARACTER',
                                                     ['atomic',
                                                      'atomic+random',
                                                      'random',
                                                      'file']],
                                     'tqr': ['LOGICAL', []]},
                       'IONS': {'bfgs_ndim': ['INTEGER', []],
                                'delta_t': ['REAL', []],
                                'ion_dynamics': ['CHARACTER',
                                                 ['bfgs',
                                                  'damp',
                                                  'verlet',
                                                  'langevin',
                                                  'langevin-smc',
                                                  'bfgs',
                                                  'damp',
                                                  'beeman']],
                                'ion_positions': ['CHARACTER', ['default', 'from_input']],
                                'ion_temperature': ['CHARACTER',
                                                    ['rescaling',
                                                     'rescale-v',
                                                     'rescale-T',
                                                     'reduce-T',
                                                     'berendsen',
                                                     'andersen',
                                                     'svr',
                                                     'initial',
                                                     'not_controlled']],
                                'ion_velocities': ['CHARACTER', ['default', 'from_input']],
                                'nraise': ['INTEGER', []],
                                'pot_extrapolation': ['CHARACTER',
                                                      ['none',
                                                       'atomic',
                                                       'first_order',
                                                       'second_order']],
                                'refold_pos': ['LOGICAL', []],
                                'remove_rigid_rot': ['LOGICAL', []],
                                'tempw': ['REAL', []],
                                'tolp': ['REAL', []],
                                'trust_radius_ini': ['REAL', []],
                                'trust_radius_max': ['REAL', []],
                                'trust_radius_min': ['REAL', []],
                                'upscale': ['REAL', []],
                                'w_1': ['REAL', []],
                                'w_2': ['REAL', []],
                                'wfc_extrapolation': ['CHARACTER',
                                                      ['none',
                                                       'first_order',
                                                       'second_order']]},
                       'SYSTEM': {'A, B, C, cosAB, cosAC, cosBC': ['REAL', []],
                                  'Hubbard_J(i,ityp), (i,ityp) = (1,1) ... (3,ntyp)': ['REAL',
                                                                                       []],
                                  'Hubbard_J0(i), i=1,ntype': ['REAL', []],
                                  'Hubbard_U(i), i=1,ntyp': ['REAL', []],
                                  'Hubbard_V(na,nb,k), (na,nb,k) = (1,1,1) ... (natx,27*natx,4)': ['REAL',
                                                                                                   []],
                                  'Hubbard_alpha(i), i=1,ntyp': ['REAL', []],
                                  'Hubbard_beta(i), i=1,ntyp': ['REAL', []],
                                  'Hubbard_parameters': ['CHARACTER', ['input', 'file']],
                                  'U_projection_type': ['CHARACTER',
                                                        ['atomic',
                                                         'ortho-atomic',
                                                         'norm-atomic',
                                                         'file',
                                                         'pseudo']],
                                  'ace': ['LOGICAL', []],
                                  'angle1(i), i=1,ntyp': ['REAL', []],
                                  'angle2(i), i=1,ntyp': ['REAL', []],
                                  'assume_isolated': ['CHARACTER',
                                                      ['none',
                                                       'makov-payne',
                                                       'm-p',
                                                       'mp',
                                                       'martyna-tuckerman',
                                                       'm-t',
                                                       'mt',
                                                       'esm',
                                                       '2D']],
                                  'block': ['LOGICAL', []],
                                  'block_1': ['REAL', []],
                                  'block_2': ['REAL', []],
                                  'block_height': ['REAL', []],
                                  'celldm(i), i=1,6': ['REAL', []],
                                  'constrained_magnetization': ['CHARACTER', []],
                                  'degauss': ['REAL', []],
                                  'dftd3_threebody': ['LOGICAL', []],
                                  'dftd3_version': ['integer',
                                                    ['dftd3_version=2',
                                                     'dftd3_version=3',
                                                     'dftd3_version=4',
                                                     'dftd3_version=5',
                                                     'dftd3_version=6']],
                                  'eamp': ['REAL', []],
                                  'ecfixed': ['REAL', []],
                                  'ecutfock': ['REAL', []],
                                  'ecutrho': ['REAL', []],
                                  'ecutvcut': ['REAL', []],
                                  'ecutwfc': ['REAL', []],
                                  'edir': ['INTEGER', []],
                                  'emaxpos': ['REAL', []],
                                  'ensemble_energies': ['LOGICAL', []],
                                  'eopreg': ['REAL', []],
                                  'esm_bc': ['CHARACTER', []],
                                  'esm_efield': ['REAL', []],
                                  'esm_nfit': ['INTEGER', []],
                                  'esm_w': ['REAL', []],
                                  'exx_fraction': ['REAL', []],
                                  'exxdiv_treatment': ['CHARACTER',
                                                       ['gygi-baldereschi',
                                                        'vcut_spherical',
                                                        'vcut_ws',
                                                        'none']],
                                  'fcp_mu': ['REAL', []],
                                  'fixed_magnetization(i), i=1,3': ['REAL', []],
                                  'force_symmorphic': ['LOGICAL', []],
                                  'ibrav': ['INTEGER', []],
                                  'input_dft': ['CHARACTER', []],
                                  'lambda': ['REAL', []],
                                  'lda_plus_u': ['LOGICAL', []],
                                  'lda_plus_u_kind': ['INTEGER', []],
                                  'lforcet': ['LOGICAL', []],
                                  'localization_thr': ['REAL', []],
                                  'london': ['LOGICAL', []],
                                  'london_c6(i), i=1,ntyp': ['REAL', []],
                                  'london_rcut': ['REAL', []],
                                  'london_rvdw(i), i=1,ntyp': ['REAL', []],
                                  'london_s6': ['REAL', []],
                                  'lspinorb': ['LOGICAL', []],
                                  'nat': ['INTEGER', []],
                                  'nbnd': ['INTEGER', []],
                                  'no_t_rev': ['LOGICAL', []],
                                  'noinv': ['LOGICAL', []],
                                  'noncolin': ['LOGICAL', []],
                                  'nosym': ['LOGICAL', []],
                                  'nosym_evc': ['LOGICAL', []],
                                  'nqx1, nqx2, nqx3': ['INTEGER', []],
                                  'nr1, nr2, nr3': ['INTEGER', []],
                                  'nr1s, nr2s, nr3s': ['INTEGER', []],
                                  'nspin': ['INTEGER', []],
                                  'ntyp': ['INTEGER', []],
                                  'occupations': ['CHARACTER',
                                                  ['smearing',
                                                   'tetrahedra',
                                                   'tetrahedra_lin',
                                                   'tetrahedra_opt',
                                                   'fixed',
                                                   'from_input']],
                                  'one_atom_occupations': ['LOGICAL', []],
                                  'origin_choice': ['INTEGER', []],
                                  'q2sigma': ['REAL', []],
                                  'qcutz': ['REAL', []],
                                  'relaxz': ['LOGICAL', []],
                                  'report': ['INTEGER', []],
                                  'rhombohedral': ['LOGICAL', []],
                                  'screening_parameter': ['REAL', []],
                                  'smearing': ['CHARACTER',
                                               ['gaussian',
                                                'gauss',
                                                'methfessel-paxton',
                                                'm-p',
                                                'mp',
                                                'marzari-vanderbilt',
                                                'cold',
                                                'm-v',
                                                'mv',
                                                'fermi-dirac',
                                                'f-d',
                                                'fd']],
                                  'space_group': ['INTEGER', []],
                                  'starting_charge(i), i=1,ntyp': ['REAL', []],
                                  'starting_magnetization(i), i=1,ntyp': ['REAL', []],
                                  'starting_ns_eigenvalue(m,ispin,ityp), (m,ispin,ityp) = (1,1,1) ... (2*lmax+1,nspin or npol,ntyp)': ['REAL',
                                                                                                                                       []],
                                  'starting_spin_angle': ['LOGICAL', []],
                                  'tot_charge': ['REAL', []],
                                  'tot_magnetization': ['REAL', []],
                                  'ts_vdw_econv_thr': ['REAL', []],
                                  'ts_vdw_isolated': ['LOGICAL', []],
                                  'uniqueb': ['LOGICAL', []],
                                  'use_all_frac': ['LOGICAL', []],
                                  'vdw_corr': ['CHARACTER', []],
                                  'x_gamma_extrapolation': ['LOGICAL', []],
                                  'xdm': ['LOGICAL', []],
                                  'xdm_a1': ['REAL', []],
                                  'xdm_a2': ['REAL', []],
                                  'zgate': ['REAL', []]},
                       'ATOMIC_FORCES': {'X': ['CHARACTER', []], 'fx, fy, fz': ['REAL', []]},
                       'ATOMIC_POSITIONS': {"Card's options:": ["Card's options:",
                                                                ['alat',
                                                                 'bohr',
                                                                 'angstrom',
                                                                 'crystal',
                                                                 'crystal_sg']],
                                            'X': ['CHARACTER', []],
                                            'if_pos(1), if_pos(2), if_pos(3)': ['INTEGER', []],
                                            'x, y, z': ['REAL', []]},
                       'ATOMIC_SPECIES': {'Mass_X': ['REAL', []],
                                          'PseudoPot_X': ['CHARACTER', []],
                                          'X': ['CHARACTER', []]},
                       'ATOMIC_VELOCITIES': {'V': ['CHARACTER', []],
                                             'vx, vy, vz': ['REAL', []]},
                       'CELL_PARAMETERS': {"Card's options:": ["Card's options:", []],
                                           'v1, v2, v3': ['REAL', []]},
                       'CONSTRAINTS': {'constr(1), constr(2), constr(3), constr(4)': ['', []],
                                       'constr_target': ['REAL', []],
                                       'constr_tol': ['REAL', []],
                                       'constr_type': ['CHARACTER',
                                                       ['type_coord',
                                                        'atom_coord',
                                                        'distance',
                                                        'planar_angle',
                                                        'torsional_angle',
                                                        'bennett_proj']],
                                       'nconstr': ['INTEGER', []]},
                       'K_POINTS': {"Card's options:": ["Card's options:",
                                                        ['tpiba',
                                                         'automatic',
                                                         'crystal',
                                                         'gamma',
                                                         'tpiba_b',
                                                         'crystal_b',
                                                         'tpiba_c',
                                                         'crystal_c']],
                                    'nk1, nk2, nk3': ['INTEGER', []],
                                    'nks': ['INTEGER', []],
                                    'sk1, sk2, sk3': ['INTEGER', []],
                                    'xk_x, xk_y, xk_z, wk': ['REAL', []]},
                       'OCCUPATIONS': {'f_inp1': ['REAL', []], 'f_inp2': ['REAL', []]}}}

default_parameters = {'BANDS': {'BANDS': {'filband': 'bands.out',
                                          'filp': 'p_avg.dat',
                                          'firstk, lastk': 1e+30,
                                          'lp': False,
                                          'lsigma(i), i=1,3': True,
                                          'lsym': True,
                                          'no_overlap': True,
                                          'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                          'currentdirectory(./)otherwise',
                                          'plot_2d': False,
                                          'prefix': 'pwscf',
                                          'spin_component': 1e+30}},
                      'DOS': {'DOS': {'DeltaE': 1e+30,
                                      'Emin, Emax': 1e+30,
                                      'bz_sum': 'smearingifdegaussisgivenininput;\n'
                                      'optionsreadfromthexmldatafileotherwise.',
                                      'degauss': 1e+30,
                                      'fildos': 'prefix.dos',
                                      'ngauss': 1e+30,
                                      'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                      'currentdirectory(./)otherwise',
                                      'prefix': 'pwscf'}},
                      'NEB': {'PATH': {'CI_scheme': 'no-CI',
                                       'ds': 1.0,
                                       'fcp_mu': 1e+30,
                                       'fcp_tot_charge_first': 1e+30,
                                       'fcp_tot_charge_last': 1e+30,
                                       'first_last_opt': False,
                                       'k_max, k_min': 1e+30,
                                       'lfcpopt': True,
                                       'minimum_image': False,
                                       'nstep_path': 1,
                                       'num_of_images': 0,
                                       'opt_scheme': 'quick-min',
                                       'path_thr': 1e+30,
                                       'restart_mode': 'from_scratch',
                                       'string_method': 'neb',
                                       'temp_req': 1e+30,
                                       'use_freezing': False,
                                       'use_masses': False},
                              'ATOMIC_POSITIONS': {},
                              'CLIMBING_IMAGES': {'index1, index2, ... indexN': 1e+30}},
                      'PP': {'INPUTPP': {'degauss_ldos': 1e+30,
                                         'delta_e': 1e+30,
                                         'emax': 1e+30,
                                         'emin': 1e+30,
                                         'filplot': '',
                                         'kband(i), i=1,2': 1e+30,
                                         'kpoint(i), i=1,2': 1e+30,
                                         'lsign': True,
                                         'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                         'currentdirectory(./)otherwise',
                                         'plot_num': 1e+30,
                                         'prefix': '',
                                         'sample_bias': 1e+30,
                                         'spin_component': 0,
                                         'spin_component(i), i=1,2': 1e+30},
                             'PLOT': {'e1(i), \n\t\te2(i), \n\t\t \n\t\t  i=1,3': 1e+30,
                                      'e1(i), \n\t\te2(i), \n\t\te3(i), \n\t\t \n\t\t  i=1,3': 1e+30,
                                      'e1(i), i=1,3': 1e+30,
                                      'fileout': 'standardoutput',
                                      'filepp(i), i=1,nfile': 'filepp(1)=filplot',
                                      'iflag': 1e+30,
                                      'interpolation': 'fourier',
                                      'nfile': 1e+30,
                                      'nx': 1e+30,
                                      'nx, ny': 1e+30,
                                      'nx, ny, nz': 1e+30,
                                      'output_format': 1e+30,
                                      'radius': 1e+30,
                                      'weight(i), i=1,nfile': 1e+30,
                                      'x0(i), i=1,3': 1e+30}},
                      'PROJWFC': {'PROJWFC': {'DeltaE': 1e+30,
                                              'Emin, Emax': 1e+30,
                                              'degauss': 0.0,
                                              'filpdos': '(valueofprefixvariable)',
                                              'filproj': '(standardoutput)',
                                              'irmax(3,n_proj_boxes)': 1e+30,
                                              'irmin(3,n_proj_boxes)': 1e+30,
                                              'kresolveddos': False,
                                              'lbinary_data': False,
                                              'lsym': True,
                                              'lwrite_overlaps': False,
                                              'n_proj_boxes': 1,
                                              'ngauss': 0,
                                              'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                              'currentdirectory(./)otherwise',
                                              'pawproj': False,
                                              'plotboxes': False,
                                              'prefix': 'pwscf',
                                              'tdosinboxes': False}},
                      'PW': {'CELL': {'cell_dofree': 'all',
                                      'cell_dynamics': '',
                                      'cell_factor': 1e+30,
                                      'press': 0.0,
                                      'press_conv_thr': 1e+30,
                                      'wmass': 1e+30},
                             'CONTROL': {'calculation': 'scf',
                                         'dipfield': False,
                                         'disk_io': 'seebelow',
                                         'dt': 20.0,
                                         'etot_conv_thr': 0.0001,
                                         'forc_conv_thr': 0.001,
                                         'gate': True,
                                         'gdir': 1e+30,
                                         'iprint': 1e+30,
                                         'lberry': False,
                                         'lelfield': False,
                                         'lfcpopt': True,
                                         'lkpoint_dir': True,
                                         'lorbm': False,
                                         'max_seconds': 1e+30,
                                         'nberrycyc': 1,
                                         'nppstr': 1e+30,
                                         'nstep': 1e+30,
                                         'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                         'currentdirectory(./)otherwise',
                                         'prefix': 'pwscf',
                                         'pseudo_dir': 'valueofthe$ESPRESSO_PSEUDOenvironmentvariableifset;\n'
                                         '$HOME/espresso/pseudo/otherwise',
                                         'restart_mode': 'from_scratch',
                                         'tefield': False,
                                         'title': '',
                                         'tprnfor': True,
                                         'tstress': False,
                                         'verbosity': 'low',
                                         'wf_collect': True,
                                         'wfcdir': 'sameasoutdir'},
                             'ELECTRONS': {'adaptive_thr': False,
                                           'conv_thr': 1e-06,
                                           'conv_thr_init': 0.001,
                                           'conv_thr_multi': 0.1,
                                           'diago_cg_maxiter': 1e+30,
                                           'diago_david_ndim': 2,
                                           'diago_full_acc': False,
                                           'diago_thr_init': 1e+30,
                                           'diagonalization': 'david',
                                           'efield': 0.0,
                                           'efield_cart(i), i=1,3': 1e+30,
                                           'efield_phase': 'none',
                                           'electron_maxstep': 100,
                                           'mixing_beta': 0.7,
                                           'mixing_fixed_ns': 0,
                                           'mixing_mode': 'plain',
                                           'mixing_ndim': 8,
                                           'real_space': False,
                                           'scf_must_converge': True,
                                           'startingpot': '',
                                           'startingwfc': 'atomic+random',
                                           'tqr': False},
                             'IONS': {'bfgs_ndim': 1,
                                      'delta_t': 1.0,
                                      'ion_dynamics': '',
                                      'ion_positions': 'default',
                                      'ion_temperature': 'not_controlled',
                                      'ion_velocities': 'default',
                                      'nraise': 1,
                                      'pot_extrapolation': 'atomic',
                                      'refold_pos': False,
                                      'remove_rigid_rot': False,
                                      'tempw': 300.0,
                                      'tolp': 100.0,
                                      'trust_radius_ini': 0.5,
                                      'trust_radius_max': 0.8,
                                      'trust_radius_min': 0.001,
                                      'upscale': 100.0,
                                      'w_1': 0.01,
                                      'w_2': 0.5,
                                      'wfc_extrapolation': 'none'},
                             'SYSTEM': {'A, B, C, cosAB, cosAC, cosBC': 1e+30,
                                        'Hubbard_J(i,ityp), (i,ityp) = (1,1) ... (3,ntyp)': 1e+30,
                                        'Hubbard_J0(i), i=1,ntype': 1e+30,
                                        'Hubbard_U(i), i=1,ntyp': 1e+30,
                                        'Hubbard_V(na,nb,k), (na,nb,k) = (1,1,1) ... (natx,27*natx,4)': 1e+30,
                                        'Hubbard_alpha(i), i=1,ntyp': 1e+30,
                                        'Hubbard_beta(i), i=1,ntyp': 1e+30,
                                        'Hubbard_parameters': 'input',
                                        'U_projection_type': 'atomic',
                                        'ace': True,
                                        'angle1(i), i=1,ntyp': 1e+30,
                                        'angle2(i), i=1,ntyp': 1e+30,
                                        'assume_isolated': 'none',
                                        'block': False,
                                        'block_1': 0.45,
                                        'block_2': 0.55,
                                        'block_height': 0.1,
                                        'celldm(i), i=1,6': 1e+30,
                                        'constrained_magnetization': '',
                                        'degauss': 1e+30,
                                        'dftd3_threebody': True,
                                        'dftd3_version': 3,
                                        'eamp': 1e+30,
                                        'ecfixed': 0.0,
                                        'ecutfock': 1e+30,
                                        'ecutrho': 1e+30,
                                        'ecutvcut': 1e+30,
                                        'ecutwfc': 1e+30,
                                        'edir': 1e+30,
                                        'emaxpos': 0.5,
                                        'ensemble_energies': False,
                                        'eopreg': 0.1,
                                        'esm_bc': '',
                                        'esm_efield': 1e+30,
                                        'esm_nfit': 1e+30,
                                        'esm_w': 1e+30,
                                        'exx_fraction': 1e+30,
                                        'exxdiv_treatment': 'gygi-baldereschi',
                                        'fcp_mu': 1e+30,
                                        'fixed_magnetization(i), i=1,3': 1e+30,
                                        'force_symmorphic': False,
                                        'ibrav': 1e+30,
                                        'input_dft': 'readfrompseudopotentialfiles',
                                        'lambda': 1e+30,
                                        'lda_plus_u': True,
                                        'lda_plus_u_kind': 0,
                                        'lforcet': True,
                                        'localization_thr': 0.0,
                                        'london': False,
                                        'london_c6(i), i=1,ntyp': 1e+30,
                                        'london_rcut': 200.0,
                                        'london_rvdw(i), i=1,ntyp': 1e+30,
                                        'london_s6': 0.75,
                                        'lspinorb': True,
                                        'nat': 1e+30,
                                        'nbnd': 1e+30,
                                        'no_t_rev': False,
                                        'noinv': False,
                                        'noncolin': False,
                                        'nosym': False,
                                        'nosym_evc': False,
                                        'nqx1, nqx2, nqx3': 1e+30,
                                        'nr1, nr2, nr3': 1e+30,
                                        'nr1s, nr2s, nr3s': 1e+30,
                                        'nspin': 1,
                                        'ntyp': 1e+30,
                                        'occupations': '',
                                        'one_atom_occupations': False,
                                        'origin_choice': 1,
                                        'q2sigma': 0.1,
                                        'qcutz': 0.0,
                                        'relaxz': False,
                                        'report': -1,
                                        'rhombohedral': True,
                                        'screening_parameter': 0.106,
                                        'smearing': 'gaussian',
                                        'space_group': 0,
                                        'starting_charge(i), i=1,ntyp': 0.0,
                                        'starting_magnetization(i), i=1,ntyp': 1e+30,
                                        'starting_ns_eigenvalue(m,ispin,ityp), (m,ispin,ityp) = (1,1,1) ... (2*lmax+1,nspin or npol,ntyp)': 1e+30,
                                        'starting_spin_angle': False,
                                        'tot_charge': 0.0,
                                        'tot_magnetization': 1e+30,
                                        'ts_vdw_econv_thr': 1e-06,
                                        'ts_vdw_isolated': False,
                                        'uniqueb': False,
                                        'use_all_frac': False,
                                        'vdw_corr': '',
                                        'x_gamma_extrapolation': True,
                                        'xdm': False,
                                        'xdm_a1': 0.6836,
                                        'xdm_a2': 1.5045,
                                        'zgate': 0.5},
                             'ATOMIC_FORCES': {'X': '', 'fx, fy, fz': 1e+30},
                             'ATOMIC_POSITIONS': {"Card's options:": '(DEPRECATED)alat',
                                                  'X': '',
                                                  'if_pos(1), if_pos(2), if_pos(3)': 1,
                                                  'x, y, z': 1e+30},
                             'ATOMIC_SPECIES': {'Mass_X': 1e+30, 'PseudoPot_X': '', 'X': ''},
                             'ATOMIC_VELOCITIES': {'V': '', 'vx, vy, vz': 1e+30},
                             'CELL_PARAMETERS': {"Card's options:": '', 'v1, v2, v3': 1e+30},
                             'CONSTRAINTS': {'constr(1), constr(2), constr(3), constr(4)': '',
                                             'constr_target': 1e+30,
                                             'constr_tol': 1e+30,
                                             'constr_type': '',
                                             'nconstr': 1e+30},
                             'K_POINTS': {"Card's options:": 'tbipa',
                                          'nk1, nk2, nk3': 1e+30,
                                          'nks': 1e+30,
                                          'sk1, sk2, sk3': 1e+30,
                                          'xk_x, xk_y, xk_z, wk': 1e+30},
                             'OCCUPATIONS': {'f_inp1': 1e+30, 'f_inp2': 1e+30}}}




from xespresso.post.base import PostCalculation


class EspressoPp(PostCalculation):

    package = 'pp'
    state_parameters = {'INPUTPP': ['prefix', 'outdir', 'filplot', 'plot_num',
                                    'spin_component', 'spin_component', 'emin', 'emax',
                                    'delta_e', 'degauss_ldos', 'sample_bias', 'kpoint',
                                    'kband', 'lsign', 'spin_component', 'emin', 'emax',
                                    'spin_component', 'spin_component', 'spin_component',
                                    'spin_component'],
                        'PLOT': ['nfile', 'filepp', 'weight', 'iflag', 'output_format',
                                 'fileout', 'interpolation', 'e1', 'x0', 'nx', 'e1', 'e2',
                                 'x0', 'nx', 'ny', 'e1', 'e2', 'e3', 'x0', 'nx', 'ny', 'nz',
                                 'radius', 'nx', 'ny']
                        }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)
